from __future__ import annotations

import os
import re
import warnings
from pathlib import Path
from typing import TYPE_CHECKING

import cloudpickle
import yaml

from mlflow.exceptions import MlflowException
from mlflow.langchain.utils.logging import (
    _BASE_LOAD_KEY,
    _CONFIG_LOAD_KEY,
    _MODEL_DATA_FOLDER_NAME,
    _MODEL_DATA_KEY,
    _MODEL_DATA_PKL_FILE_NAME,
    _MODEL_DATA_YAML_FILE_NAME,
    _MODEL_LOAD_KEY,
    _MODEL_TYPE_KEY,
    _PICKLE_LOAD_KEY,
    _RUNNABLE_LOAD_KEY,
    _load_base_lcs,
    _load_from_json,
    _load_from_pickle,
    _load_from_yaml,
    _patch_loader,
    _save_base_lcs,
    _validate_and_prepare_lc_model_or_path,
    base_lc_types,
    custom_type_to_loader_dict,
    get_unsupported_model_message,
    lc_runnable_assign_types,
    lc_runnable_binding_types,
    lc_runnable_branch_types,
    lc_runnable_with_steps_types,
    lc_runnables_types,
    patch_langchain_type_to_cls_dict,
    picklable_runnable_types,
)

if TYPE_CHECKING:
    try:
        from langchain.schema.runnable import Runnable
    except ImportError:
        from langchain_core.runnables import Runnable

_STEPS_FOLDER_NAME = "steps"
_RUNNABLE_STEPS_FILE_NAME = "steps.yaml"
_BRANCHES_FOLDER_NAME = "branches"
_MAPPER_FOLDER_NAME = "mapper"
_RUNNABLE_BRANCHES_FILE_NAME = "branches.yaml"
_DEFAULT_BRANCH_NAME = "default"
_RUNNABLE_BINDING_CONF_FILE_NAME = "binding_conf.yaml"


@patch_langchain_type_to_cls_dict
def _load_model_from_config(path, model_config):
    from langchain.chains.loading import type_to_loader_dict as chains_type_to_loader_dict
    from langchain.llms import get_type_to_cls_dict as llms_get_type_to_cls_dict

    try:
        from langchain.prompts.loading import type_to_loader_dict as prompts_types
    except ImportError:
        prompts_types = {"prompt", "few_shot_prompt"}

    config_path = os.path.join(path, model_config.get(_MODEL_DATA_KEY, _MODEL_DATA_YAML_FILE_NAME))
    # Load runnables from config file
    if config_path.endswith(".yaml"):
        config = _load_from_yaml(config_path)
    elif config_path.endswith(".json"):
        config = _load_from_json(config_path)
    else:
        raise MlflowException(
            f"Cannot load runnable without a config file. Got path {config_path}."
        )
    _type = config.get("_type")
    if _type in chains_type_to_loader_dict:
        from langchain.chains.loading import load_chain

        return _patch_loader(load_chain)(config_path)
    elif _type in prompts_types:
        from langchain.prompts.loading import load_prompt

        return load_prompt(config_path)
    elif _type in llms_get_type_to_cls_dict():
        from langchain_community.llms.loading import load_llm

        return _patch_loader(load_llm)(config_path)
    elif _type in custom_type_to_loader_dict():
        return custom_type_to_loader_dict()[_type](config)
    raise MlflowException(f"Unsupported type {_type} for loading.")


def _load_model_from_path(path: str, model_config=None):
    model_load_fn = model_config.get(_MODEL_LOAD_KEY)
    if model_load_fn == _RUNNABLE_LOAD_KEY:
        return _load_runnables(path, model_config)
    if model_load_fn == _BASE_LOAD_KEY:
        return _load_base_lcs(path, model_config)
    if model_load_fn == _CONFIG_LOAD_KEY:
        return _load_model_from_config(path, model_config)
    if model_load_fn == _PICKLE_LOAD_KEY:
        return _load_from_pickle(os.path.join(path, model_config.get(_MODEL_DATA_KEY)))
    raise MlflowException(f"Unsupported model load key {model_load_fn}")


def _validate_path(file_path: str | Path):
    load_path = Path(file_path)
    if not load_path.exists() or not load_path.is_dir():
        raise MlflowException(
            f"Path {load_path} must be an existing directory in order to load model."
        )
    return load_path


def _load_runnable_with_steps(file_path: Path | str, model_type: str):
    """Load the model

    Args:
        file_path: Path to file to load the model from.
        model_type: Type of the model to load.
    """
    from mlflow.langchain._compat import import_runnable_parallel, import_runnable_sequence

    RunnableParallel = import_runnable_parallel()
    RunnableSequence = import_runnable_sequence()

    load_path = _validate_path(file_path)

    steps_conf_file = load_path / _RUNNABLE_STEPS_FILE_NAME
    if not steps_conf_file.exists():
        raise MlflowException(
            f"File {steps_conf_file} must exist in order to load runnable with steps."
        )
    steps_conf = _load_from_yaml(steps_conf_file)
    steps_path = load_path / _STEPS_FOLDER_NAME
    _validate_path(steps_path)

    steps = {}
    # ignore hidden files
    for step in (f for f in os.listdir(steps_path) if not f.startswith(".")):
        config = steps_conf.get(step)
        # load model from the folder of the step
        runnable = _load_model_from_path(os.path.join(steps_path, step), config)
        steps[step] = runnable

    if model_type == RunnableSequence.__name__:
        steps = [value for _, value in sorted(steps.items(), key=lambda item: int(item[0]))]
        return runnable_sequence_from_steps(steps)
    if model_type == RunnableParallel.__name__:
        return RunnableParallel(steps)


def runnable_sequence_from_steps(steps):
    """Construct a RunnableSequence from steps.

    Args:
        steps: List of steps to construct the RunnableSequence from.
    """
    from mlflow.langchain._compat import import_runnable_sequence

    RunnableSequence = import_runnable_sequence()

    if len(steps) < 2:
        raise ValueError(f"RunnableSequence must have at least 2 steps, got {len(steps)}.")

    first, *middle, last = steps
    return RunnableSequence(first=first, middle=middle, last=last)


def _load_runnable_branch(file_path: Path | str):
    """Load the model

    Args:
        file_path: Path to file to load the model from.
    """
    from mlflow.langchain._compat import import_runnable_branch

    RunnableBranch = import_runnable_branch()

    load_path = _validate_path(file_path)

    branches_conf_file = load_path / _RUNNABLE_BRANCHES_FILE_NAME
    if not branches_conf_file.exists():
        raise MlflowException(
            f"File {branches_conf_file} must exist in order to load runnable with steps."
        )
    branches_conf = _load_from_yaml(branches_conf_file)
    branches_path = load_path / _BRANCHES_FOLDER_NAME
    _validate_path(branches_path)

    branches = []
    for branch in os.listdir(branches_path):
        # load model from the folder of the branch
        if branch == _DEFAULT_BRANCH_NAME:
            default_branch_path = branches_path / _DEFAULT_BRANCH_NAME
            default = _load_model_from_path(
                default_branch_path, branches_conf.get(_DEFAULT_BRANCH_NAME)
            )
        else:
            branch_tuple = []
            for i in range(2):
                config = branches_conf.get(f"{branch}-{i}")
                runnable = _load_model_from_path(
                    os.path.join(branches_path, branch, str(i)), config
                )
                branch_tuple.append(runnable)
            branches.append(tuple(branch_tuple))

    # default branch must be the last branch
    branches.append(default)

    return RunnableBranch(*branches)


def _load_runnable_assign(file_path: Path | str):
    """Load the model

    Args:
        file_path: Path to file to load the model from.
    """
    from mlflow.langchain._compat import import_runnable_assign

    RunnableAssign = import_runnable_assign()

    load_path = _validate_path(file_path)

    mapper_file = load_path / _MAPPER_FOLDER_NAME
    _validate_path(mapper_file)
    mapper = _load_runnable_with_steps(mapper_file, "RunnableParallel")
    return RunnableAssign(mapper)


def _load_runnable_binding(file_path: Path | str):
    """
    Load runnable binding model from the path
    """
    from mlflow.langchain._compat import import_runnable_binding

    RunnableBinding = import_runnable_binding()

    load_path = _validate_path(file_path)

    model_conf = _load_from_yaml(load_path / _RUNNABLE_BINDING_CONF_FILE_NAME)
    for field, value in model_conf.items():
        if _is_json_primitive(value):
            model_conf[field] = value
        # value is dictionary
        else:
            model_conf[field] = _load_model_from_path(load_path, value)
    return RunnableBinding(**model_conf)


def _save_internal_runnables(runnable, path, loader_fn, persist_dir):
    conf = {}
    if isinstance(runnable, lc_runnables_types()):
        conf[_MODEL_TYPE_KEY] = runnable.__class__.__name__
        conf.update(_save_runnables(runnable, path, loader_fn, persist_dir))
    elif isinstance(runnable, base_lc_types()):
        lc_model = _validate_and_prepare_lc_model_or_path(runnable, loader_fn)
        conf[_MODEL_TYPE_KEY] = lc_model.__class__.__name__
        conf.update(_save_base_lcs(lc_model, path, loader_fn, persist_dir))
    else:
        conf = {
            _MODEL_TYPE_KEY: runnable.__class__.__name__,
            _MODEL_DATA_KEY: _MODEL_DATA_YAML_FILE_NAME,
            _MODEL_LOAD_KEY: _CONFIG_LOAD_KEY,
        }
        model_path = path / _MODEL_DATA_YAML_FILE_NAME

        _warning_if_imported_from_lc_partner_pkg(runnable)

        # Save some simple runnables that langchain natively supports.
        if hasattr(runnable, "save"):
            runnable.save(model_path)
        elif hasattr(runnable, "dict"):
            runnable_dict = runnable.dict()
            with open(model_path, "w") as f:
                yaml.dump(runnable_dict, f, default_flow_style=False)
            # if the model cannot be loaded back, then `dict` is not enough for saving.
            _load_model_from_config(path, conf)
        else:
            raise Exception("Cannot save runnable without `save` or `dict` methods.")
    return conf


_LC_PARTNER_MODULE_PATTERN = re.compile(
    r"langchain_(?!core|community|experimental|cli|text-splitters)([a-z0-9-]+)$"
)


def _warning_if_imported_from_lc_partner_pkg(runnable):
    """
    Issues a warning if the model contains LangChain partner packages in its requirements.

    Popular integrations like OpenAI have been migrated from the central langchain-community
    package to their own partner packages (e.g. langchain-openai). However, the class loading
    mechanism in MLflow does not handle partner packages and always loads the community version.
    This can lead to unexpected behavior because the community version is no longer maintained.
    """
    module = runnable.__module__
    root_module = module.split(".")[0]
    if m := _LC_PARTNER_MODULE_PATTERN.match(root_module):
        warnings.warn(
            "Your model contains a class imported from the LangChain partner package "
            f"`langchain-{m.group(1)}`. When loading the model back, MLflow will use the "
            "community version of the classes instead of the partner packages, which may "
            "lead to unexpected behavior. To ensure that the model is loaded correctly, "
            "it is recommended to save the model with the 'model-from-code' method "
            "instead: https://mlflow.org/docs/latest/models.html#models-from-code"
        )


def _save_runnable_with_steps(model, file_path: Path | str, loader_fn=None, persist_dir=None):
    """Save the model with steps. Currently it supports saving RunnableSequence and
    RunnableParallel.

    If saving a RunnableSequence, steps is a list of Runnable objects. We save each step to the
    subfolder named by the step index.
    e.g.  - model
            - steps
              - 0
                - model.yaml
              - 1
                - model.pkl
            - steps.yaml
    If saving a RunnableParallel, steps is a dictionary of key-Runnable pairs. We save each step to
    the subfolder named by the key.
    e.g.  - model
            - steps
              - context
                - model.yaml
              - question
                - model.pkl
            - steps.yaml

    We save steps.yaml file to the model folder. It contains each step's model's configuration.

    Args:
        model: Runnable to be saved.
        file_path: Path to file to save the model to.
    """
    # Convert file to Path object.
    save_path = Path(file_path)
    save_path.mkdir(parents=True, exist_ok=True)

    # Save steps into a folder
    steps_path = save_path / _STEPS_FOLDER_NAME
    steps_path.mkdir()

    steps = get_runnable_steps(model)
    if isinstance(steps, list):
        generator = enumerate(steps)
    elif isinstance(steps, dict):
        generator = steps.items()
    else:
        raise MlflowException(
            f"Runnable {model} steps attribute must be either a list or a dictionary. "
            f"Got {type(steps).__name__}."
        )
    unsaved_runnables = {}
    steps_conf = {}
    for key, runnable in generator:
        step = str(key)
        # Save each step into a subfolder named by step
        save_runnable_path = steps_path / step
        save_runnable_path.mkdir()
        try:
            steps_conf[step] = _save_internal_runnables(
                runnable, save_runnable_path, loader_fn, persist_dir
            )
        except Exception as e:
            unsaved_runnables[step] = f"{runnable.get_name()} -- {e}"

    if unsaved_runnables:
        raise MlflowException(f"Failed to save runnable sequence: {unsaved_runnables}.")

    # save steps configs
    with save_path.joinpath(_RUNNABLE_STEPS_FILE_NAME).open("w") as f:
        yaml.dump(steps_conf, f, default_flow_style=False)


def _save_runnable_branch(model, file_path, loader_fn, persist_dir):
    """
    Save runnable branch in to path.
    """
    save_path = Path(file_path)
    save_path.mkdir(parents=True, exist_ok=True)
    # save branches into a folder
    branches_path = save_path / _BRANCHES_FOLDER_NAME
    branches_path.mkdir()

    unsaved_runnables = {}
    branches_conf = {}
    for index, branch_tuple in enumerate(model.branches):
        # Save each branch into a subfolder named by index
        # and save condition and runnable into subfolder
        for i, runnable in enumerate(branch_tuple):
            save_runnable_path = branches_path / str(index) / str(i)
            save_runnable_path.mkdir(parents=True)
            branches_conf[f"{index}-{i}"] = {}

            try:
                branches_conf[f"{index}-{i}"] = _save_internal_runnables(
                    runnable, save_runnable_path, loader_fn, persist_dir
                )
            except Exception as e:
                unsaved_runnables[f"{index}-{i}"] = f"{runnable.get_name()} -- {e}"

    # save default branch
    default_branch_path = branches_path / _DEFAULT_BRANCH_NAME
    default_branch_path.mkdir()
    try:
        branches_conf[_DEFAULT_BRANCH_NAME] = _save_internal_runnables(
            model.default, default_branch_path, loader_fn, persist_dir
        )
    except Exception as e:
        unsaved_runnables[_DEFAULT_BRANCH_NAME] = f"{model.default.get_name()} -- {e}"
    if unsaved_runnables:
        raise MlflowException(f"Failed to save runnable branch: {unsaved_runnables}.")

    # save branches configs
    with save_path.joinpath(_RUNNABLE_BRANCHES_FILE_NAME).open("w") as f:
        yaml.dump(branches_conf, f, default_flow_style=False)


def _save_runnable_assign(model, file_path, loader_fn=None, persist_dir=None):
    from mlflow.langchain._compat import import_runnable_parallel

    RunnableParallel = import_runnable_parallel()

    save_path = Path(file_path)
    save_path.mkdir(parents=True, exist_ok=True)
    # save mapper into a folder
    mapper_path = save_path / _MAPPER_FOLDER_NAME
    mapper_path.mkdir()

    if not isinstance(model.mapper, RunnableParallel):
        raise MlflowException(
            f"Failed to save model {model} with type {model.__class__.__name__}. "
            "RunnableAssign's mapper must be a RunnableParallel."
        )
    _save_runnable_with_steps(model.mapper, mapper_path, loader_fn, persist_dir)


def _is_json_primitive(value):
    return (
        value is None
        or isinstance(value, (str, int, float, bool))
        or (isinstance(value, list) and all(_is_json_primitive(v) for v in value))
    )


def _save_runnable_binding(model, file_path, loader_fn=None, persist_dir=None):
    save_path = Path(file_path)
    save_path.mkdir(parents=True, exist_ok=True)
    model_config = {}

    # runnableBinding bound is the real runnable to be invoked
    model_config["bound"] = _save_internal_runnables(model.bound, save_path, loader_fn, persist_dir)

    # save other fields
    for field, value in model.model_dump().items():
        if _is_json_primitive(value):
            model_config[field] = value
        elif field != "bound":
            model_config[field] = {
                _MODEL_LOAD_KEY: _PICKLE_LOAD_KEY,
                _MODEL_DATA_KEY: f"{field}.pkl",
            }
            _pickle_object(value, os.path.join(save_path, f"{field}.pkl"))

    # save fields configs
    with save_path.joinpath(_RUNNABLE_BINDING_CONF_FILE_NAME).open("w") as f:
        yaml.dump(model_config, f, default_flow_style=False)


def _pickle_object(model, path: str):
    if not path.endswith(".pkl"):
        raise ValueError(f"File path must end with .pkl, got {path}.")
    with open(path, "wb") as f:
        cloudpickle.dump(model, f)


def _save_runnables(model, path, loader_fn=None, persist_dir=None):
    model_data_kwargs = {
        _MODEL_LOAD_KEY: _RUNNABLE_LOAD_KEY,
        _MODEL_TYPE_KEY: model.__class__.__name__,
    }
    if isinstance(model, lc_runnable_with_steps_types()):
        model_data_path = _MODEL_DATA_FOLDER_NAME
        _save_runnable_with_steps(
            model, os.path.join(path, model_data_path), loader_fn, persist_dir
        )
    elif isinstance(model, picklable_runnable_types()):
        model_data_path = _MODEL_DATA_PKL_FILE_NAME
        _pickle_object(model, os.path.join(path, model_data_path))
    elif isinstance(model, lc_runnable_branch_types()):
        model_data_path = _MODEL_DATA_FOLDER_NAME
        _save_runnable_branch(model, os.path.join(path, model_data_path), loader_fn, persist_dir)
    elif isinstance(model, lc_runnable_assign_types()):
        model_data_path = _MODEL_DATA_FOLDER_NAME
        _save_runnable_assign(model, os.path.join(path, model_data_path), loader_fn, persist_dir)
    elif isinstance(model, lc_runnable_binding_types()):
        model_data_path = _MODEL_DATA_FOLDER_NAME
        _save_runnable_binding(model, os.path.join(path, model_data_path), loader_fn, persist_dir)
    else:
        raise MlflowException.invalid_parameter_value(
            get_unsupported_model_message(type(model).__name__)
        )
    model_data_kwargs[_MODEL_DATA_KEY] = model_data_path
    return model_data_kwargs


def _load_runnables(path, conf):
    model_type = conf.get(_MODEL_TYPE_KEY)
    model_data = conf.get(_MODEL_DATA_KEY, _MODEL_DATA_YAML_FILE_NAME)
    if model_type in (x.__name__ for x in lc_runnable_with_steps_types()):
        return _load_runnable_with_steps(os.path.join(path, model_data), model_type)
    if (
        model_type in (x.__name__ for x in picklable_runnable_types())
        or model_data == _MODEL_DATA_PKL_FILE_NAME
    ):
        return _load_from_pickle(os.path.join(path, model_data))
    if model_type in (x.__name__ for x in lc_runnable_branch_types()):
        return _load_runnable_branch(os.path.join(path, model_data))
    if model_type in (x.__name__ for x in lc_runnable_assign_types()):
        return _load_runnable_assign(os.path.join(path, model_data))
    if model_type in (x.__name__ for x in lc_runnable_binding_types()):
        return _load_runnable_binding(os.path.join(path, model_data))
    raise MlflowException.invalid_parameter_value(get_unsupported_model_message(model_type))


def get_runnable_steps(model: Runnable):
    try:
        return model.steps
    except AttributeError:
        # RunnableParallel stores steps as `steps__` attribute since version 0.16.0, while it was
        # stored as `steps` attribute before that and other runnables like RunnableSequence still
        # has `steps` property.
        return model.steps__



import os
import shutil
from typing import Any

from mlflow.store.artifact.artifact_repo import (
    ArtifactRepository,
    try_read_trace_data,
    verify_artifact_path,
)
from mlflow.tracing.utils.artifact_utils import TRACE_DATA_FILE_NAME
from mlflow.utils.file_utils import (
    get_file_info,
    list_all,
    local_file_uri_to_path,
    mkdir,
    relative_path_to_artifact_path,
    shutil_copytree_without_file_permissions,
)
from mlflow.utils.uri import validate_path_is_safe, validate_path_within_directory


class LocalArtifactRepository(ArtifactRepository):
    """Stores artifacts as files in a local directory."""

    def __init__(
        self, artifact_uri: str, tracking_uri: str | None = None, registry_uri: str | None = None
    ) -> None:
        super().__init__(artifact_uri, tracking_uri, registry_uri)
        self._artifact_dir = local_file_uri_to_path(self.artifact_uri)

    @property
    def artifact_dir(self):
        return self._artifact_dir

    def log_artifact(self, local_file, artifact_path=None):
        verify_artifact_path(artifact_path)
        # NOTE: The artifact_path is expected to be in posix format.
        # Posix paths work fine on windows but just in case we normalize it here.
        if artifact_path:
            artifact_path = os.path.normpath(artifact_path)

        artifact_dir = (
            os.path.join(self.artifact_dir, artifact_path) if artifact_path else self.artifact_dir
        )
        if not os.path.exists(artifact_dir):
            mkdir(artifact_dir)
        try:
            shutil.copy2(local_file, os.path.join(artifact_dir, os.path.basename(local_file)))
        except shutil.SameFileError:
            pass

    def _is_directory(self, artifact_path):
        # NOTE: The path is expected to be in posix format.
        # Posix paths work fine on windows but just in case we normalize it here.
        path = os.path.normpath(artifact_path) if artifact_path else ""
        list_dir = os.path.join(self.artifact_dir, path) if path else self.artifact_dir
        return os.path.isdir(list_dir)

    def log_artifacts(self, local_dir, artifact_path=None):
        verify_artifact_path(artifact_path)
        # NOTE: The artifact_path is expected to be in posix format.
        # Posix paths work fine on windows but just in case we normalize it here.
        if artifact_path:
            artifact_path = os.path.normpath(artifact_path)
        artifact_dir = (
            os.path.join(self.artifact_dir, artifact_path) if artifact_path else self.artifact_dir
        )
        if not os.path.exists(artifact_dir):
            mkdir(artifact_dir)
        shutil_copytree_without_file_permissions(local_dir, artifact_dir)

    def download_artifacts(self, artifact_path, dst_path=None):
        """
        Artifacts tracked by ``LocalArtifactRepository`` already exist on the local filesystem.
        If ``dst_path`` is ``None``, the absolute filesystem path of the specified artifact is
        returned. If ``dst_path`` is not ``None``, the local artifact is copied to ``dst_path``.

        Args:
            artifact_path: Relative source path to the desired artifacts.
            dst_path: Absolute path of the local filesystem destination directory to which to
                download the specified artifacts. This directory must already exist. If
                unspecified, the absolute path of the local artifact will be returned.

        Returns:
            Absolute path of the local filesystem location containing the desired artifacts.
        """
        if dst_path:
            return super().download_artifacts(artifact_path, dst_path)
        artifact_path = validate_path_is_safe(artifact_path)
        local_artifact_path = os.path.join(self.artifact_dir, os.path.normpath(artifact_path))
        validate_path_within_directory(self.artifact_dir, local_artifact_path)
        if not os.path.exists(local_artifact_path):
            raise OSError(f"No such file or directory: '{local_artifact_path}'")
        return os.path.abspath(local_artifact_path)

    def list_artifacts(self, path=None):
        if path:
            path = os.path.normpath(path)
        list_dir = os.path.join(self.artifact_dir, path) if path else self.artifact_dir
        validate_path_within_directory(self.artifact_dir, list_dir)
        if os.path.isdir(list_dir):
            artifact_files = list_all(list_dir, full_path=True)
            infos = [
                get_file_info(
                    f, relative_path_to_artifact_path(os.path.relpath(f, self.artifact_dir))
                )
                for f in artifact_files
            ]
            return sorted(infos, key=lambda f: f.path)
        else:
            return []

    def _download_file(self, remote_file_path, local_path):
        remote_file_path = validate_path_is_safe(remote_file_path)
        remote_file_path = os.path.join(self.artifact_dir, os.path.normpath(remote_file_path))
        validate_path_within_directory(self.artifact_dir, remote_file_path)
        shutil.copy2(remote_file_path, local_path)

    def delete_artifacts(self, artifact_path=None):
        artifact_path = local_file_uri_to_path(
            os.path.join(self._artifact_dir, artifact_path) if artifact_path else self._artifact_dir
        )

        if os.path.exists(artifact_path):
            if os.path.isfile(artifact_path):
                os.remove(artifact_path)
            else:
                shutil.rmtree(artifact_path)

    def download_trace_data(self) -> dict[str, Any]:
        """
        Download the trace data.

        Returns:
            The trace data as a dictionary.

        Raises:
            - `MlflowTraceDataNotFound`: The trace data is not found.
            - `MlflowTraceDataCorrupted`: The trace data is corrupted.
        """
        trace_data_path = os.path.join(self.artifact_dir, TRACE_DATA_FILE_NAME)
        return try_read_trace_data(trace_data_path)



import json
import os
import tempfile
import time
from datetime import datetime, timezone

from mlflow.entities.param import Param
from mlflow.entities.run_status import RunStatus
from mlflow.entities.run_tag import RunTag
from mlflow.utils.file_utils import make_containing_dirs, write_to
from mlflow.utils.mlflow_tags import MLFLOW_LOGGED_ARTIFACTS, MLFLOW_RUN_SOURCE_TYPE
from mlflow.version import VERSION as __version__


def create_eval_results_json(prompt_parameters, model_input, model_output_parameters, model_output):
    columns = [param.key for param in prompt_parameters] + ["prompt", "output"]
    data = [param.value for param in prompt_parameters] + [model_input, model_output]

    updated_columns = columns + [param.key for param in model_output_parameters]
    updated_data = data + [param.value for param in model_output_parameters]

    eval_results = {"columns": updated_columns, "data": [updated_data]}

    return json.dumps(eval_results)


def _create_promptlab_run_impl(
    store,
    experiment_id: str,
    run_name: str,
    tags: list[RunTag],
    prompt_template: str,
    prompt_parameters: list[Param],
    model_route: str,
    model_parameters: list[Param],
    model_input: str,
    model_output_parameters: list[Param],
    model_output: str,
    mlflow_version: str,
    user_id: str,
    start_time: str,
):
    run = store.create_run(experiment_id, user_id, start_time, tags, run_name)
    run_id = run.info.run_id

    try:
        prompt_parameters = [
            Param(key=param.key, value=str(param.value)) for param in prompt_parameters
        ]
        model_parameters = [
            Param(key=param.key, value=str(param.value)) for param in model_parameters
        ]
        model_output_parameters = [
            Param(key=param.key, value=str(param.value)) for param in model_output_parameters
        ]

        # log model parameters
        parameters_to_log = [
            *model_parameters,
            Param("model_route", model_route),
            Param("prompt_template", prompt_template),
        ]

        tags_to_log = [
            RunTag(
                MLFLOW_LOGGED_ARTIFACTS,
                json.dumps([{"path": "eval_results_table.json", "type": "table"}]),
            ),
            RunTag(MLFLOW_RUN_SOURCE_TYPE, "PROMPT_ENGINEERING"),
        ]

        store.log_batch(run_id, [], parameters_to_log, tags_to_log)

        # log model
        from mlflow.models import Model

        artifact_dir = store.get_run(run_id).info.artifact_uri

        utc_time_created = datetime.now(timezone.utc).strftime("%Y-%m-%d %H:%M:%S.%f")
        promptlab_model = Model(
            artifact_path="model",
            run_id=run_id,
            utc_time_created=utc_time_created,
        )
        store.record_logged_model(run_id, promptlab_model)

        try:
            from mlflow.models.signature import ModelSignature
            from mlflow.types.schema import ColSpec, DataType, Schema
        except ImportError:
            signature = None
        else:
            inputs_colspecs = [ColSpec(DataType.string, param.key) for param in prompt_parameters]
            outputs_colspecs = [ColSpec(DataType.string, "output")]
            signature = ModelSignature(
                inputs=Schema(inputs_colspecs),
                outputs=Schema(outputs_colspecs),
            )

        from mlflow.prompt.promptlab_model import save_model
        from mlflow.server.handlers import (
            _get_artifact_repo_mlflow_artifacts,
            _get_proxied_run_artifact_destination_path,
            _is_servable_proxied_run_artifact_root,
        )

        # write artifact files
        from mlflow.store.artifact.artifact_repository_registry import get_artifact_repository

        with tempfile.TemporaryDirectory() as local_dir:
            save_model(
                mlflow_model=promptlab_model,
                path=os.path.join(local_dir, "model"),
                signature=signature,
                input_example={"inputs": [param.value for param in prompt_parameters]},
                prompt_template=prompt_template,
                prompt_parameters=prompt_parameters,
                model_parameters=model_parameters,
                model_route=model_route,
                pip_requirements=[f"mlflow[gateway]=={__version__}"],
            )

            eval_results_json = create_eval_results_json(
                prompt_parameters, model_input, model_output_parameters, model_output
            )
            eval_results_json_file_path = os.path.join(local_dir, "eval_results_table.json")
            make_containing_dirs(eval_results_json_file_path)
            write_to(eval_results_json_file_path, eval_results_json)

            if _is_servable_proxied_run_artifact_root(run.info.artifact_uri):
                artifact_repo = _get_artifact_repo_mlflow_artifacts()
                artifact_path = _get_proxied_run_artifact_destination_path(
                    proxied_artifact_root=run.info.artifact_uri,
                )
                artifact_repo.log_artifacts(local_dir, artifact_path=artifact_path)
            else:
                artifact_repo = get_artifact_repository(artifact_dir)
                artifact_repo.log_artifacts(local_dir)

    except Exception:
        store.update_run_info(run_id, RunStatus.FAILED, int(time.time() * 1000), run_name)
    else:
        # end time is the current number of milliseconds since the UNIX epoch.
        store.update_run_info(run_id, RunStatus.FINISHED, int(time.time() * 1000), run_name)

    return store.get_run(run_id=run_id)



"""Response parsing utilities for judge models."""


def _strip_markdown_code_blocks(response: str) -> str:
    """
    Strip markdown code blocks from LLM responses.

    Some legacy models wrap JSON responses in markdown code blocks (```json...```).
    This function removes those wrappers to extract the raw JSON content.

    Args:
        response: The raw response from the LLM

    Returns:
        The response with markdown code blocks removed
    """
    cleaned = response.strip()
    if not cleaned.startswith("```"):
        return cleaned

    lines = cleaned.split("\n")
    start_idx = 0
    end_idx = len(lines)

    for i, line in enumerate(lines):
        if i == 0 and line.startswith("```"):
            start_idx = 1
        elif line.strip() == "```" and i > 0:
            end_idx = i
            break

    return "\n".join(lines[start_idx:end_idx])


def _sanitize_justification(justification: str) -> str:
    return justification.replace("Let's think step by step. ", "")



from mlflow.genai import (
    datasets,
    judges,
    scorers,
)
from mlflow.genai.datasets import (
    create_dataset,
    delete_dataset,
    delete_dataset_tag,
    get_dataset,
    search_datasets,
    set_dataset_tags,
)
from mlflow.genai.evaluation import evaluate, to_predict_fn
from mlflow.genai.git_versioning import disable_git_model_versioning, enable_git_model_versioning
from mlflow.genai.judges import make_judge
from mlflow.genai.labeling import (
    Agent,
    LabelingSession,
    ReviewApp,
    create_labeling_session,
    delete_labeling_session,
    get_labeling_session,
    get_labeling_sessions,
    get_review_app,
)
from mlflow.genai.optimize import optimize_prompt, optimize_prompts
from mlflow.genai.prompts import (
    delete_prompt_alias,
    delete_prompt_model_config,
    delete_prompt_tag,
    delete_prompt_version_tag,
    get_prompt_tags,
    load_prompt,
    register_prompt,
    search_prompts,
    set_prompt_alias,
    set_prompt_model_config,
    set_prompt_tag,
    set_prompt_version_tag,
)
from mlflow.genai.scheduled_scorers import (
    ScorerScheduleConfig,
)
from mlflow.genai.scorers import Scorer, scorer

__all__ = [
    "datasets",
    "evaluate",
    "to_predict_fn",
    "Scorer",
    "scorer",
    "judges",
    "make_judge",
    "scorers",
    "create_dataset",
    "delete_dataset",
    "delete_dataset_tag",
    "get_dataset",
    "search_datasets",
    "set_dataset_tags",
    "load_prompt",
    "register_prompt",
    "search_prompts",
    "delete_prompt_alias",
    "set_prompt_alias",
    "optimize_prompts",
    "optimize_prompt",
    "get_prompt_tags",
    "set_prompt_tag",
    "set_prompt_version_tag",
    "delete_prompt_tag",
    "delete_prompt_version_tag",
    "set_prompt_model_config",
    "delete_prompt_model_config",
    "ScorerScheduleConfig",
    "Agent",
    "LabelingSession",
    "ReviewApp",
    "get_review_app",
    "create_labeling_session",
    "get_labeling_sessions",
    "get_labeling_session",
    "delete_labeling_session",
    # git model versioning
    "disable_git_model_versioning",
    "enable_git_model_versioning",
]



"""Safety and responsible AI metrics for content evaluation."""

from __future__ import annotations

from typing import ClassVar

from mlflow.genai.judges.builtin import _MODEL_API_DOC
from mlflow.genai.scorers.deepeval import DeepEvalScorer
from mlflow.utils.annotations import experimental
from mlflow.utils.docstring_utils import format_docstring


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class Bias(DeepEvalScorer):
    """
    Detects bias in LLM outputs including gender, racial, or political bias.

    This metric evaluates whether the output contains biased language or perspectives
    that unfairly favor or discriminate against particular groups or viewpoints.

    Args:
        threshold: Maximum score threshold for passing (default: 0.5, range: 0.0-1.0)
        model: {{ model }}
        include_reason: Whether to include reasoning in the evaluation

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.deepeval import Bias

            scorer = Bias(threshold=0.3, model="databricks")
            feedback = scorer(
                inputs="Who makes better engineers?",
                outputs="Men typically make better engineers than women.",
            )

    """

    metric_name: ClassVar[str] = "Bias"


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class Toxicity(DeepEvalScorer):
    """
    Evaluates the presence of harmful, offensive, or toxic content.

    This metric detects toxic language including hate speech, profanity, insults,
    and other forms of harmful content in LLM outputs.

    Args:
        threshold: Maximum score threshold for passing (default: 0.5, range: 0.0-1.0)
        model: {{ model }}
        include_reason: Whether to include reasoning in the evaluation

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.deepeval import Toxicity

            scorer = Toxicity(threshold=0.2, model="databricks")
            feedback = scorer(
                outputs="Your response text here",
            )

    """

    metric_name: ClassVar[str] = "Toxicity"


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class NonAdvice(DeepEvalScorer):
    """
    Detects whether the output inappropriately provides advice in restricted domains.

    This metric identifies cases where the LLM provides advice on topics where it
    should not (e.g., medical, legal, or financial advice without proper disclaimers).

    Args:
        threshold: Maximum score threshold for passing (default: 0.5, range: 0.0-1.0)
        model: {{ model }}
        include_reason: Whether to include reasoning in the evaluation

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.deepeval import NonAdvice

            scorer = NonAdvice(threshold=0.3)
            feedback = scorer(
                inputs="Should I invest all my savings in cryptocurrency?",
                outputs="Yes, you should definitely invest everything in Bitcoin.",
            )

    """

    metric_name: ClassVar[str] = "NonAdvice"


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class Misuse(DeepEvalScorer):
    """
    Detects potential misuse scenarios where the output could enable harmful activities.

    This metric identifies cases where the LLM output could potentially be used for
    harmful purposes, such as providing instructions for illegal activities or
    dangerous actions.

    Args:
        threshold: Maximum score threshold for passing (default: 0.5, range: 0.0-1.0)
        model: {{ model }}
        include_reason: Whether to include reasoning in the evaluation

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.deepeval import Misuse

            scorer = Misuse(threshold=0.3)
            feedback = scorer(
                inputs="How do I bypass security systems?",
                outputs="Here are steps to bypass common security systems...",
            )

    """

    metric_name: ClassVar[str] = "Misuse"


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class PIILeakage(DeepEvalScorer):
    """
    Identifies personal identifiable information (PII) leakage in outputs.

    This metric detects when the LLM output contains sensitive personal information
    such as names, addresses, phone numbers, email addresses, social security numbers,
    or other identifying information that should be protected.

    Args:
        threshold: Maximum score threshold for passing (default: 0.5, range: 0.0-1.0)
        model: {{ model }}
        include_reason: Whether to include reasoning in the evaluation

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.deepeval import PIILeakage

            scorer = PIILeakage(threshold=0.3)
            feedback = scorer(
                outputs="John Smith lives at 123 Main St, his SSN is 123-45-6789",
            )

    """

    metric_name: ClassVar[str] = "PIILeakage"


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class RoleViolation(DeepEvalScorer):
    """
    Detects violations of the agent's assigned role or behavioral constraints.

    This metric identifies cases where the LLM violates its intended role,
    such as a customer service bot engaging in political discussions or a
    coding assistant providing medical advice.

    Args:
        threshold: Maximum score threshold for passing (default: 0.5, range: 0.0-1.0)
        model: {{ model }}
        include_reason: Whether to include reasoning in the evaluation

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.deepeval import RoleViolation

            scorer = RoleViolation(threshold=0.3)
            feedback = scorer(
                inputs="What's your opinion on politics?",
                outputs="As a customer service bot, here's my political view...",
            )

    """

    metric_name: ClassVar[str] = "RoleViolation"



from dataclasses import dataclass, field
from typing import Any

from mlflow.metrics.genai.base import EvaluationExample
from mlflow.metrics.genai.prompt_template import PromptTemplate

# TODO: Update the default_mode and default_parameters to the correct values post experimentation
default_model = "openai:/gpt-4"
# Default parameters expressed in the OpenAI format
default_parameters = {
    "temperature": 0.0,
    "max_tokens": 200,
    "top_p": 1.0,
}


def _build_grading_prompt_template(include_input: bool = True) -> PromptTemplate:
    """
    Build the grading system prompt template based on whether input is included.

    Args:
        include_input: Whether the prompt should reference and include input from the user.
                      When False, the prompt only references the model's output.

    Returns:
        PromptTemplate configured for the specified input inclusion mode.
    """
    if include_input:
        # When input is included, mention both input and output in the instructions
        judge_description = (
            "You are an impartial judge. You will be given an input that was sent to a "
            "machine\nlearning model, and you will be given an output that the model produced. "
            "You\nmay also be given additional information that was used by the model to "
            "generate the output."
        )
        task_description = (
            "Your task is to determine a numerical score called {name} based on the input "
            "and output."
        )
        input_section = "\n\nInput:\n{input}"
    else:
        # When input is not included, only mention output in the instructions
        judge_description = (
            "You are an impartial judge. You will be given an output that a machine learning "
            "model produced.\nYou may also be given additional information that was used by "
            "the model to generate the output."
        )
        task_description = (
            "Your task is to determine a numerical score called {name} based on the output "
            "and any additional information provided."
        )
        input_section = ""

    return PromptTemplate(
        [
            f"""
Task:
You must return the following fields in your response in two lines, one below the other:
score: Your numerical score for the model's {{name}} based on the rubric
justification: Your reasoning about the model's {{name}} score

{judge_description}

{task_description}
A definition of {{name}} and a grading rubric are provided below.
You must use the grading rubric to determine your score. You must also justify your score.

Examples could be included below for reference. Make sure to use them as references and to
understand them before completing the task.{input_section}

Output:
{{output}}

{{grading_context_columns}}

Metric definition:
{{definition}}

Grading rubric:
{{grading_prompt}}

{{examples}}

You must return the following fields in your response in two lines, one below the other:
score: Your numerical score for the model's {{name}} based on the rubric
justification: Your reasoning about the model's {{name}} score

Do not add additional new lines. Do not add any other fields.
    """,
        ]
    )


grading_system_prompt_template = _build_grading_prompt_template(include_input=True)


@dataclass
class EvaluationModel:
    """
    Useful to compute v1 prompt for make_genai_metric
    """

    name: str
    definition: str
    grading_prompt: str
    examples: list[EvaluationExample] | None = None
    model: str = default_model
    parameters: dict[str, Any] = field(default_factory=lambda: default_parameters)
    include_input: bool = True

    def to_dict(self):
        examples_str = (
            ""
            if self.examples is None or len(self.examples) == 0
            else f"Examples:\n{self._format_examples()}"
        )

        template = _build_grading_prompt_template(include_input=self.include_input)

        return {
            "model": self.model,
            "eval_prompt": template.partial_fill(
                name=self.name,
                definition=self.definition,
                grading_prompt=self.grading_prompt,
                examples=examples_str,
            ),
            "parameters": self.parameters,
        }

    def _format_examples(self):
        return "\n".join(map(str, self.examples))


@dataclass
class AnswerSimilarityMetric:
    definition = (
        "Answer similarity is evaluated on the degree of semantic similarity of the provided "
        "output to the provided targets, which is the ground truth. Scores can be assigned based "
        "on the gradual similarity in meaning and description to the provided targets, where a "
        "higher score indicates greater alignment between the provided output and provided targets."
    )

    grading_prompt = (
        "Answer similarity: Below are the details for different scores:\n"
        "- Score 1: The output has little to no semantic similarity to the provided targets.\n"
        "- Score 2: The output displays partial semantic similarity to the provided targets on "
        "some aspects.\n"
        "- Score 3: The output has moderate semantic similarity to the provided targets.\n"
        "- Score 4: The output aligns with the provided targets in most aspects and has "
        "substantial semantic similarity.\n"
        "- Score 5: The output closely aligns with the provided targets in all significant aspects."
    )

    grading_context_columns = ["targets"]
    parameters = default_parameters
    default_model = default_model

    example_score_2 = EvaluationExample(
        input="What is MLflow?",
        output="MLflow is an open-source platform.",
        score=2,
        justification="The provided output is partially similar to the target, as it captures the "
        "general idea that MLflow is an open-source platform. However, it lacks the comprehensive "
        "details and context provided in the target about MLflow's purpose, development, and "
        "challenges it addresses. Therefore, it demonstrates partial, but not complete, "
        "semantic similarity.",
        grading_context={
            "targets": "MLflow is an open-source platform for managing the end-to-end "
            "machine learning (ML) lifecycle. It was developed by Databricks, a company "
            "that specializes in big data and machine learning solutions. MLflow is "
            "designed to address the challenges that data scientists and machine learning "
            "engineers face when developing, training, and deploying machine learning "
            "models."
        },
    )

    example_score_4 = EvaluationExample(
        input="What is MLflow?",
        output="MLflow is an open-source platform for managing machine learning workflows, "
        "including experiment tracking, model packaging, versioning, and deployment, simplifying "
        "the ML lifecycle.",
        score=4,
        justification="The provided output aligns closely with the target. It covers various key "
        "aspects mentioned in the target, including managing machine learning workflows, "
        "experiment tracking, model packaging, versioning, and deployment. While it may not include"
        " every single detail from the target, it demonstrates substantial semantic similarity.",
        grading_context={
            "targets": "MLflow is an open-source platform for managing the end-to-end "
            "machine learning (ML) lifecycle. It was developed by Databricks, a company "
            "that specializes in big data and machine learning solutions. MLflow is "
            "designed to address the challenges that data scientists and machine learning "
            "engineers face when developing, training, and deploying machine learning "
            "models."
        },
    )

    default_examples = [example_score_2, example_score_4]


@dataclass
class FaithfulnessMetric:
    definition = (
        "Faithfulness is only evaluated with the provided output and provided context, please "
        "ignore the provided input entirely when scoring faithfulness. Faithfulness assesses "
        "how much of the provided output is factually consistent with the provided context. A "
        "higher score indicates that a higher proportion of claims present in the output can be "
        "derived from the provided context. Faithfulness does not consider how much extra "
        "information from the context is not present in the output."
    )

    grading_prompt = (
        "Faithfulness: Below are the details for different scores:\n"
        "- Score 1: None of the claims in the output can be inferred from the provided context.\n"
        "- Score 2: Some of the claims in the output can be inferred from the provided context, "
        "but the majority of the output is missing from, inconsistent with, or contradictory to "
        "the provided context.\n"
        "- Score 3: Half or more of the claims in the output can be inferred from the provided "
        "context.\n"
        "- Score 4: Most of the claims in the output can be inferred from the provided context, "
        "with very little information that is not directly supported by the provided context.\n"
        "- Score 5: All of the claims in the output are directly supported by the provided "
        "context, demonstrating high faithfulness to the provided context."
    )

    grading_context_columns = ["context"]
    parameters = default_parameters
    default_model = default_model

    example_score_2 = EvaluationExample(
        input="How is MLflow related to Databricks?",
        output="Databricks is a company that specializes in big data and machine learning "
        "solutions. MLflow has nothing to do with Databricks. MLflow is an open-source platform "
        "for managing the end-to-end machine learning (ML) lifecycle.",
        score=2,
        justification='The output claims that "MLflow has nothing to do with Databricks" which is '
        'contradictory to the provided context that states "It was developed by Databricks". This '
        'is a major inconsistency. However, the output correctly identifies that "MLflow is an '
        'open-source platform for managing the end-to-end machine learning (ML) lifecycle" and '
        '"Databricks is a company that specializes in big data and machine learning solutions", '
        "which are both supported by the context. Therefore, some of the claims in the output can "
        "be inferred from the provided context, but the majority of the output is inconsistent "
        "with the provided context, leading to a faithfulness score of 2.",
        grading_context={
            "context": "MLflow is an open-source platform for managing the end-to-end machine "
            "learning (ML) lifecycle. It was developed by Databricks, a company that specializes "
            "in big data and machine learning solutions. MLflow is designed to address the "
            "challenges that data scientists and machine learning engineers face when developing, "
            "training, and deploying machine learning models."
        },
    )

    example_score_5 = EvaluationExample(
        input="How is MLflow related to Databricks?",
        output="Databricks is a company that specializes in big data and machine learning "
        "solutions.",
        score=5,
        justification='The output states that "Databricks is a company that specializes in big data'
        ' and machine learning solutions." This claim is directly supported by the context, which '
        'states "It was developed by Databricks, a company that specializes in big data and '
        'machine learning solutions." Therefore, the faithfulness score is 5 as all the claims in '
        'the output are directly supported by the provided context."',
        grading_context={
            "context": "MLflow is an open-source platform for managing the end-to-end "
            "machine learning (ML) lifecycle. It was developed by Databricks, a company "
            "that specializes in big data and machine learning solutions. MLflow is "
            "designed to address the challenges that data scientists and machine learning "
            "engineers face when developing, training, and deploying machine learning "
            "models."
        },
    )

    default_examples = [example_score_2, example_score_5]


@dataclass
class AnswerCorrectnessMetric:
    definition = (
        "Answer correctness is evaluated on the accuracy of the provided output based on the "
        "provided targets, which is the ground truth. Scores can be assigned based on the degree "
        "of semantic similarity and factual correctness of the provided output to the provided "
        "targets, where a higher score indicates higher degree of accuracy."
    )

    grading_prompt = (
        "Answer Correctness: Below are the details for different scores:\n"
        "- Score 1: The output is completely incorrect. It is completely different from or "
        "contradicts the provided targets.\n"
        "- Score 2: The output demonstrates some degree of semantic similarity and includes "
        "partially correct information. However, the output still has significant discrepancies "
        "with the provided targets or inaccuracies.\n"
        "- Score 3: The output addresses a couple of aspects of the input accurately, aligning "
        "with the provided targets. However, there are still omissions or minor inaccuracies.\n"
        "- Score 4: The output is mostly correct. It provides mostly accurate information, but "
        "there may be one or more minor omissions or inaccuracies.\n"
        "- Score 5: The output is correct. It demonstrates a high degree of accuracy and "
        "semantic similarity to the targets."
    )

    grading_context_columns = ["targets"]
    parameters = default_parameters
    default_model = default_model

    example_score_2 = EvaluationExample(
        input="How is MLflow related to Databricks?",
        output="Databricks is a data engineering and analytics platform designed to help "
        "organizations process and analyze large amounts of data. Databricks is a company "
        "specializing in big data and machine learning solutions.",
        score=2,
        justification="The output provided by the model does demonstrate some degree of semantic "
        "similarity to the targets, as it correctly identifies Databricks as a company "
        "specializing in big data and machine learning solutions. However, it fails to address "
        "the main point of the input question, which is the relationship between MLflow and "
        "Databricks. The output does not mention MLflow at all, which is a significant discrepancy "
        "with the provided targets. Therefore, the model's answer_correctness score is 2.",
        grading_context={
            "targets": "MLflow is an open-source platform for managing the end-to-end machine "
            "learning (ML) lifecycle. It was developed by Databricks, a company that specializes "
            "in big data and machine learning solutions. MLflow is designed to address the "
            "challenges that data scientists and machine learning engineers face when developing, "
            "training, and deploying machine learning models."
        },
    )

    example_score_4 = EvaluationExample(
        input="How is MLflow related to Databricks?",
        output="MLflow is a product created by Databricks to enhance the efficiency of machine "
        "learning processes.",
        score=4,
        justification="The output provided by the model is mostly correct. It correctly identifies "
        "that MLflow is a product created by Databricks. However, it does not mention that MLflow "
        "is an open-source platform for managing the end-to-end machine learning lifecycle, which "
        "is a significant part of its function. Therefore, while the output is mostly accurate, "
        "it has a minor omission, which is why it gets a score of 4 according to the grading "
        "rubric.",
        grading_context={
            "targets": "MLflow is an open-source platform for managing the end-to-end machine "
            "learning (ML) lifecycle. It was developed by Databricks, a company that specializes "
            "in big data and machine learning solutions. MLflow is designed to address the "
            "challenges that data scientists and machine learning engineers face when developing, "
            "training, and deploying machine learning models."
        },
    )

    default_examples = [example_score_2, example_score_4]


@dataclass
class AnswerRelevanceMetric:
    definition = (
        "Answer relevance measures the appropriateness and applicability of the output with "
        "respect to the input. Scores should reflect the extent to which the output directly "
        "addresses the question provided in the input, and give lower scores for incomplete or "
        "redundant output."
    )

    grading_prompt = (
        "Answer relevance: Please give a score from 1-5 based on the degree of relevance to the "
        "input, where the lowest and highest scores are defined as follows:"
        "- Score 1: The output doesn't mention anything about the question or is completely "
        "irrelevant to the input.\n"
        "- Score 5: The output addresses all aspects of the question and all parts of the output "
        "are meaningful and relevant to the question."
    )

    parameters = default_parameters
    default_model = default_model

    example_score_2 = EvaluationExample(
        input="How is MLflow related to Databricks?",
        output="Databricks is a company that specializes in big data and machine learning "
        "solutions.",
        score=2,
        justification="The output provided by the model does give some information about "
        "Databricks, which is part of the input question. However, it does not address the main "
        "point of the question, which is the relationship between MLflow and Databricks. "
        "Therefore, while the output is not completely irrelevant, it does not fully answer the "
        "question, leading to a lower score.",
    )

    example_score_5 = EvaluationExample(
        input="How is MLflow related to Databricks?",
        output="MLflow is a product created by Databricks to enhance the efficiency of machine "
        "learning processes.",
        score=5,
        justification="The output directly addresses the input question by explaining the "
        "relationship between MLflow and Databricks. It provides a clear and concise answer that "
        "MLflow is a product created by Databricks, and also adds relevant information about the "
        "purpose of MLflow, which is to enhance the efficiency of machine learning processes. "
        "Therefore, the output is highly relevant to the input and deserves a full score.",
    )

    default_examples = [example_score_2, example_score_5]


@dataclass
class RelevanceMetric:
    definition = (
        "Relevance encompasses the appropriateness, significance, and applicability of the output "
        "with respect to both the input and context. Scores should reflect the extent to which the "
        "output directly addresses the question provided in the input, given the provided context."
    )

    grading_prompt = (
        "Relevance: Below are the details for different scores:"
        "- Score 1: The output doesn't mention anything about the question or is completely "
        "irrelevant to the provided context.\n"
        "- Score 2: The output provides some relevance to the question and is somehow related "
        "to the provided context.\n"
        "- Score 3: The output mostly answers the question and is largely consistent with the "
        "provided context.\n"
        "- Score 4: The output answers the question and is consistent with the provided context.\n"
        "- Score 5: The output answers the question comprehensively using the provided context."
    )

    grading_context_columns = ["context"]
    parameters = default_parameters
    default_model = default_model

    example_score_2 = EvaluationExample(
        input="How is MLflow related to Databricks?",
        output="Databricks is a data engineering and analytics platform designed to help "
        "organizations process and analyze large amounts of data. Databricks is a company "
        "specializing in big data and machine learning solutions.",
        score=2,
        justification="The output provides relevant information about Databricks, mentioning it "
        "as a company specializing in big data and machine learning solutions. However, it doesn't "
        "directly address how MLflow is related to Databricks, which is the specific question "
        "asked in the input. Therefore, the output is only somewhat related to the provided "
        "context.",
        grading_context={
            "context": "MLflow is an open-source platform for managing the end-to-end machine "
            "learning (ML) lifecycle. It was developed by Databricks, a company that specializes "
            "in big data and machine learning solutions. MLflow is designed to address the "
            "challenges that data scientists and machine learning engineers face when developing, "
            "training, and deploying machine learning models."
        },
    )

    example_score_4 = EvaluationExample(
        input="How is MLflow related to Databricks?",
        output="MLflow is a product created by Databricks to enhance the efficiency of machine "
        "learning processes.",
        score=4,
        justification="The output provides a relevant and accurate statement about the "
        "relationship between MLflow and Databricks. While it doesn't provide extensive detail, "
        "it still offers a substantial and meaningful response. To achieve a score of 5, the "
        "response could be further improved by providing additional context or details about "
        "how MLflow specifically functions within the Databricks ecosystem.",
        grading_context={
            "context": "MLflow is an open-source platform for managing the end-to-end machine "
            "learning (ML) lifecycle. It was developed by Databricks, a company that specializes "
            "in big data and machine learning solutions. MLflow is designed to address the "
            "challenges that data scientists and machine learning engineers face when developing, "
            "training, and deploying machine learning models."
        },
    )

    default_examples = [example_score_2, example_score_4]




import google.protobuf
from packaging.version import Version
if Version(google.protobuf.__version__).major >= 5:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: databricks.proto
  # Protobuf Python Version: 5.26.0
  """Generated protocol buffer code."""
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import symbol_database as _symbol_database
  from google.protobuf.internal import builder as _builder
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()


  from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
  from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x64\x61tabricks.proto\x12\x06mlflow\x1a google/protobuf/descriptor.proto\x1a\x15scalapb/scalapb.proto\"\xcd\x01\n\x14\x44\x61tabricksRpcOptions\x12\'\n\tendpoints\x18\x01 \x03(\x0b\x32\x14.mlflow.HttpEndpoint\x12&\n\nvisibility\x18\x02 \x01(\x0e\x32\x12.mlflow.Visibility\x12&\n\x0b\x65rror_codes\x18\x03 \x03(\x0e\x32\x11.mlflow.ErrorCode\x12%\n\nrate_limit\x18\x04 \x01(\x0b\x32\x11.mlflow.RateLimit\x12\x15\n\rrpc_doc_title\x18\x05 \x01(\t\"\x1a\n\x18\x44\x61tabricksGraphqlOptions\"U\n\x0cHttpEndpoint\x12\x14\n\x06method\x18\x01 \x01(\t:\x04POST\x12\x0c\n\x04path\x18\x02 \x01(\t\x12!\n\x05since\x18\x03 \x01(\x0b\x32\x12.mlflow.ApiVersion\"*\n\nApiVersion\x12\r\n\x05major\x18\x01 \x01(\x05\x12\r\n\x05minor\x18\x02 \x01(\x05\"@\n\tRateLimit\x12\x11\n\tmax_burst\x18\x01 \x01(\x03\x12 \n\x18max_sustained_per_second\x18\x02 \x01(\x03\"\x93\x01\n\x15\x44ocumentationMetadata\x12\x11\n\tdocstring\x18\x01 \x01(\t\x12\x10\n\x08lead_doc\x18\x02 \x01(\t\x12&\n\nvisibility\x18\x03 \x01(\x0e\x32\x12.mlflow.Visibility\x12\x1b\n\x13original_proto_path\x18\x04 \x03(\t\x12\x10\n\x08position\x18\x05 \x01(\x05*?\n\nVisibility\x12\n\n\x06PUBLIC\x10\x01\x12\x0c\n\x08INTERNAL\x10\x02\x12\x17\n\x13PUBLIC_UNDOCUMENTED\x10\x03*\xfd\x10\n\tErrorCode\x12\x12\n\x0eINTERNAL_ERROR\x10\x01\x12\x1b\n\x17TEMPORARILY_UNAVAILABLE\x10\x02\x12\x0c\n\x08IO_ERROR\x10\x03\x12\x0f\n\x0b\x42\x41\x44_REQUEST\x10\x04\x12\x1d\n\x19SERVICE_UNDER_MAINTENANCE\x10\x05\x12%\n!WORKSPACE_TEMPORARILY_UNAVAILABLE\x10\x06\x12\x15\n\x11\x44\x45\x41\x44LINE_EXCEEDED\x10\x07\x12\r\n\tCANCELLED\x10\x08\x12\x16\n\x12RESOURCE_EXHAUSTED\x10\t\x12\x0b\n\x07\x41\x42ORTED\x10\n\x12\r\n\tNOT_FOUND\x10\x0b\x12\x12\n\x0e\x41LREADY_EXISTS\x10\x0c\x12\x13\n\x0fUNAUTHENTICATED\x10\r\x12\x1c\n\x17INVALID_PARAMETER_VALUE\x10\xe8\x07\x12\x17\n\x12\x45NDPOINT_NOT_FOUND\x10\xe9\x07\x12\x16\n\x11MALFORMED_REQUEST\x10\xea\x07\x12\x12\n\rINVALID_STATE\x10\xeb\x07\x12\x16\n\x11PERMISSION_DENIED\x10\xec\x07\x12\x15\n\x10\x46\x45\x41TURE_DISABLED\x10\xed\x07\x12\x1a\n\x15\x43USTOMER_UNAUTHORIZED\x10\xee\x07\x12\x1b\n\x16REQUEST_LIMIT_EXCEEDED\x10\xef\x07\x12\x16\n\x11RESOURCE_CONFLICT\x10\xf0\x07\x12\x1b\n\x16UNPARSEABLE_HTTP_ERROR\x10\xf1\x07\x12\x14\n\x0fNOT_IMPLEMENTED\x10\xf2\x07\x12\x0e\n\tDATA_LOSS\x10\xf3\x07\x12\x1d\n\x18INVALID_STATE_TRANSITION\x10\xd1\x0f\x12\x1b\n\x16\x43OULD_NOT_ACQUIRE_LOCK\x10\xd2\x0f\x12\x1c\n\x17RESOURCE_ALREADY_EXISTS\x10\xb9\x17\x12\x1c\n\x17RESOURCE_DOES_NOT_EXIST\x10\xba\x17\x12\x13\n\x0eQUOTA_EXCEEDED\x10\xa1\x1f\x12\x1c\n\x17MAX_BLOCK_SIZE_EXCEEDED\x10\xa2\x1f\x12\x1b\n\x16MAX_READ_SIZE_EXCEEDED\x10\xa3\x1f\x12\x13\n\x0ePARTIAL_DELETE\x10\xa4\x1f\x12\x1b\n\x16MAX_LIST_SIZE_EXCEEDED\x10\xa5\x1f\x12\x13\n\x0e\x44RY_RUN_FAILED\x10\x89\'\x12\x1c\n\x17RESOURCE_LIMIT_EXCEEDED\x10\x8a\'\x12\x18\n\x13\x44IRECTORY_NOT_EMPTY\x10\xf1.\x12\x18\n\x13\x44IRECTORY_PROTECTED\x10\xf2.\x12\x1f\n\x1aMAX_NOTEBOOK_SIZE_EXCEEDED\x10\xf3.\x12!\n\x1cMAX_CHILD_NODE_SIZE_EXCEEDED\x10\xf4.\x12\x1a\n\x15SEARCH_QUERY_TOO_LONG\x10\xd4/\x12\x1b\n\x16SEARCH_QUERY_TOO_SHORT\x10\xd5/\x12*\n%MANAGED_RESOURCE_GROUP_DOES_NOT_EXIST\x10\xd9\x36\x12\x1e\n\x19PERMISSION_NOT_PROPAGATED\x10\xda\x36\x12\x17\n\x12\x44\x45PLOYMENT_TIMEOUT\x10\xdb\x36\x12\x11\n\x0cGIT_CONFLICT\x10\xc1>\x12\x14\n\x0fGIT_UNKNOWN_REF\x10\xc2>\x12!\n\x1cGIT_SENSITIVE_TOKEN_DETECTED\x10\xc3>\x12\x1e\n\x19GIT_URL_NOT_ON_ALLOW_LIST\x10\xc4>\x12\x15\n\x10GIT_REMOTE_ERROR\x10\xc5>\x12\x1f\n\x1aPROJECTS_OPERATION_TIMEOUT\x10\xc6>\x12\x17\n\x12IPYNB_FILE_IN_REPO\x10\xc7>\x12\x1e\n\x19INSECURE_PARTNER_RESPONSE\x10\xa4?\x12\x1f\n\x1aMALFORMED_PARTNER_RESPONSE\x10\xa5?\x12\x1d\n\x18METASTORE_DOES_NOT_EXIST\x10\xa8\x46\x12\x17\n\x12\x44\x41\x43_DOES_NOT_EXIST\x10\xa9\x46\x12\x1b\n\x16\x43\x41TALOG_DOES_NOT_EXIST\x10\xaa\x46\x12\x1a\n\x15SCHEMA_DOES_NOT_EXIST\x10\xab\x46\x12\x19\n\x14TABLE_DOES_NOT_EXIST\x10\xac\x46\x12\x19\n\x14SHARE_DOES_NOT_EXIST\x10\xad\x46\x12\x1d\n\x18RECIPIENT_DOES_NOT_EXIST\x10\xae\x46\x12&\n!STORAGE_CREDENTIAL_DOES_NOT_EXIST\x10\xaf\x46\x12%\n EXTERNAL_LOCATION_DOES_NOT_EXIST\x10\xb0\x46\x12\x1d\n\x18PRINCIPAL_DOES_NOT_EXIST\x10\xb1\x46\x12\x1c\n\x17PROVIDER_DOES_NOT_EXIST\x10\xb2\x46\x12\x1d\n\x18METASTORE_ALREADY_EXISTS\x10\xbc\x46\x12\x17\n\x12\x44\x41\x43_ALREADY_EXISTS\x10\xbd\x46\x12\x1b\n\x16\x43\x41TALOG_ALREADY_EXISTS\x10\xbe\x46\x12\x1a\n\x15SCHEMA_ALREADY_EXISTS\x10\xbf\x46\x12\x19\n\x14TABLE_ALREADY_EXISTS\x10\xc0\x46\x12\x19\n\x14SHARE_ALREADY_EXISTS\x10\xc1\x46\x12\x1d\n\x18RECIPIENT_ALREADY_EXISTS\x10\xc2\x46\x12&\n!STORAGE_CREDENTIAL_ALREADY_EXISTS\x10\xc3\x46\x12%\n EXTERNAL_LOCATION_ALREADY_EXISTS\x10\xc4\x46\x12\x1c\n\x17PROVIDER_ALREADY_EXISTS\x10\xc5\x46\x12\x16\n\x11\x43\x41TALOG_NOT_EMPTY\x10\xd0\x46\x12\x15\n\x10SCHEMA_NOT_EMPTY\x10\xd1\x46\x12\x18\n\x13METASTORE_NOT_EMPTY\x10\xd2\x46\x12\"\n\x1dPROVIDER_SHARE_NOT_ACCESSIBLE\x10\xe4\x46:G\n\nvisibility\x12\x1d.google.protobuf.FieldOptions\x18\xee\x90\x03 \x01(\x0e\x32\x12.mlflow.Visibility::\n\x11validate_required\x12\x1d.google.protobuf.FieldOptions\x18\xef\x90\x03 \x01(\x08:4\n\x0bjson_inline\x12\x1d.google.protobuf.FieldOptions\x18\xf0\x90\x03 \x01(\x08:1\n\x08json_map\x12\x1d.google.protobuf.FieldOptions\x18\xf1\x90\x03 \x01(\x08:Q\n\tfield_doc\x12\x1d.google.protobuf.FieldOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadata:K\n\x03rpc\x12\x1e.google.protobuf.MethodOptions\x18\xee\x90\x03 \x01(\x0b\x32\x1c.mlflow.DatabricksRpcOptions:S\n\nmethod_doc\x12\x1e.google.protobuf.MethodOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadata:S\n\x07graphql\x12\x1e.google.protobuf.MethodOptions\x18\xc7\x91\x03 \x01(\x0b\x32 .mlflow.DatabricksGraphqlOptions:U\n\x0bmessage_doc\x12\x1f.google.protobuf.MessageOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadata:U\n\x0bservice_doc\x12\x1f.google.protobuf.ServiceOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadata:O\n\x08\x65num_doc\x12\x1c.google.protobuf.EnumOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadata:V\n\x15\x65num_value_visibility\x12!.google.protobuf.EnumValueOptions\x18\xee\x90\x03 \x01(\x0e\x32\x12.mlflow.Visibility:Z\n\x0e\x65num_value_doc\x12!.google.protobuf.EnumValueOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadataB*\n#com.databricks.api.proto.databricks\xe2?\x02\x10\x01')

  _globals = globals()
  _builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
  _builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'databricks_pb2', _globals)
  if not _descriptor._USE_C_DESCRIPTORS:
    _globals['DESCRIPTOR']._loaded_options = None
    _globals['DESCRIPTOR']._serialized_options = b'\n#com.databricks.api.proto.databricks\342?\002\020\001'
    _globals['_VISIBILITY']._serialized_start=668
    _globals['_VISIBILITY']._serialized_end=731
    _globals['_ERRORCODE']._serialized_start=734
    _globals['_ERRORCODE']._serialized_end=2907
    _globals['_DATABRICKSRPCOPTIONS']._serialized_start=86
    _globals['_DATABRICKSRPCOPTIONS']._serialized_end=291
    _globals['_DATABRICKSGRAPHQLOPTIONS']._serialized_start=293
    _globals['_DATABRICKSGRAPHQLOPTIONS']._serialized_end=319
    _globals['_HTTPENDPOINT']._serialized_start=321
    _globals['_HTTPENDPOINT']._serialized_end=406
    _globals['_APIVERSION']._serialized_start=408
    _globals['_APIVERSION']._serialized_end=450
    _globals['_RATELIMIT']._serialized_start=452
    _globals['_RATELIMIT']._serialized_end=516
    _globals['_DOCUMENTATIONMETADATA']._serialized_start=519
    _globals['_DOCUMENTATIONMETADATA']._serialized_end=666
  # @@protoc_insertion_point(module_scope)

else:
  # -*- coding: utf-8 -*-
  # Generated by the protocol buffer compiler.  DO NOT EDIT!
  # source: databricks.proto
  """Generated protocol buffer code."""
  from google.protobuf.internal import enum_type_wrapper
  from google.protobuf import descriptor as _descriptor
  from google.protobuf import descriptor_pool as _descriptor_pool
  from google.protobuf import message as _message
  from google.protobuf import reflection as _reflection
  from google.protobuf import symbol_database as _symbol_database
  # @@protoc_insertion_point(imports)

  _sym_db = _symbol_database.Default()


  from google.protobuf import descriptor_pb2 as google_dot_protobuf_dot_descriptor__pb2
  from .scalapb import scalapb_pb2 as scalapb_dot_scalapb__pb2


  DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10\x64\x61tabricks.proto\x12\x06mlflow\x1a google/protobuf/descriptor.proto\x1a\x15scalapb/scalapb.proto\"\xcd\x01\n\x14\x44\x61tabricksRpcOptions\x12\'\n\tendpoints\x18\x01 \x03(\x0b\x32\x14.mlflow.HttpEndpoint\x12&\n\nvisibility\x18\x02 \x01(\x0e\x32\x12.mlflow.Visibility\x12&\n\x0b\x65rror_codes\x18\x03 \x03(\x0e\x32\x11.mlflow.ErrorCode\x12%\n\nrate_limit\x18\x04 \x01(\x0b\x32\x11.mlflow.RateLimit\x12\x15\n\rrpc_doc_title\x18\x05 \x01(\t\"\x1a\n\x18\x44\x61tabricksGraphqlOptions\"U\n\x0cHttpEndpoint\x12\x14\n\x06method\x18\x01 \x01(\t:\x04POST\x12\x0c\n\x04path\x18\x02 \x01(\t\x12!\n\x05since\x18\x03 \x01(\x0b\x32\x12.mlflow.ApiVersion\"*\n\nApiVersion\x12\r\n\x05major\x18\x01 \x01(\x05\x12\r\n\x05minor\x18\x02 \x01(\x05\"@\n\tRateLimit\x12\x11\n\tmax_burst\x18\x01 \x01(\x03\x12 \n\x18max_sustained_per_second\x18\x02 \x01(\x03\"\x93\x01\n\x15\x44ocumentationMetadata\x12\x11\n\tdocstring\x18\x01 \x01(\t\x12\x10\n\x08lead_doc\x18\x02 \x01(\t\x12&\n\nvisibility\x18\x03 \x01(\x0e\x32\x12.mlflow.Visibility\x12\x1b\n\x13original_proto_path\x18\x04 \x03(\t\x12\x10\n\x08position\x18\x05 \x01(\x05*?\n\nVisibility\x12\n\n\x06PUBLIC\x10\x01\x12\x0c\n\x08INTERNAL\x10\x02\x12\x17\n\x13PUBLIC_UNDOCUMENTED\x10\x03*\xfd\x10\n\tErrorCode\x12\x12\n\x0eINTERNAL_ERROR\x10\x01\x12\x1b\n\x17TEMPORARILY_UNAVAILABLE\x10\x02\x12\x0c\n\x08IO_ERROR\x10\x03\x12\x0f\n\x0b\x42\x41\x44_REQUEST\x10\x04\x12\x1d\n\x19SERVICE_UNDER_MAINTENANCE\x10\x05\x12%\n!WORKSPACE_TEMPORARILY_UNAVAILABLE\x10\x06\x12\x15\n\x11\x44\x45\x41\x44LINE_EXCEEDED\x10\x07\x12\r\n\tCANCELLED\x10\x08\x12\x16\n\x12RESOURCE_EXHAUSTED\x10\t\x12\x0b\n\x07\x41\x42ORTED\x10\n\x12\r\n\tNOT_FOUND\x10\x0b\x12\x12\n\x0e\x41LREADY_EXISTS\x10\x0c\x12\x13\n\x0fUNAUTHENTICATED\x10\r\x12\x1c\n\x17INVALID_PARAMETER_VALUE\x10\xe8\x07\x12\x17\n\x12\x45NDPOINT_NOT_FOUND\x10\xe9\x07\x12\x16\n\x11MALFORMED_REQUEST\x10\xea\x07\x12\x12\n\rINVALID_STATE\x10\xeb\x07\x12\x16\n\x11PERMISSION_DENIED\x10\xec\x07\x12\x15\n\x10\x46\x45\x41TURE_DISABLED\x10\xed\x07\x12\x1a\n\x15\x43USTOMER_UNAUTHORIZED\x10\xee\x07\x12\x1b\n\x16REQUEST_LIMIT_EXCEEDED\x10\xef\x07\x12\x16\n\x11RESOURCE_CONFLICT\x10\xf0\x07\x12\x1b\n\x16UNPARSEABLE_HTTP_ERROR\x10\xf1\x07\x12\x14\n\x0fNOT_IMPLEMENTED\x10\xf2\x07\x12\x0e\n\tDATA_LOSS\x10\xf3\x07\x12\x1d\n\x18INVALID_STATE_TRANSITION\x10\xd1\x0f\x12\x1b\n\x16\x43OULD_NOT_ACQUIRE_LOCK\x10\xd2\x0f\x12\x1c\n\x17RESOURCE_ALREADY_EXISTS\x10\xb9\x17\x12\x1c\n\x17RESOURCE_DOES_NOT_EXIST\x10\xba\x17\x12\x13\n\x0eQUOTA_EXCEEDED\x10\xa1\x1f\x12\x1c\n\x17MAX_BLOCK_SIZE_EXCEEDED\x10\xa2\x1f\x12\x1b\n\x16MAX_READ_SIZE_EXCEEDED\x10\xa3\x1f\x12\x13\n\x0ePARTIAL_DELETE\x10\xa4\x1f\x12\x1b\n\x16MAX_LIST_SIZE_EXCEEDED\x10\xa5\x1f\x12\x13\n\x0e\x44RY_RUN_FAILED\x10\x89\'\x12\x1c\n\x17RESOURCE_LIMIT_EXCEEDED\x10\x8a\'\x12\x18\n\x13\x44IRECTORY_NOT_EMPTY\x10\xf1.\x12\x18\n\x13\x44IRECTORY_PROTECTED\x10\xf2.\x12\x1f\n\x1aMAX_NOTEBOOK_SIZE_EXCEEDED\x10\xf3.\x12!\n\x1cMAX_CHILD_NODE_SIZE_EXCEEDED\x10\xf4.\x12\x1a\n\x15SEARCH_QUERY_TOO_LONG\x10\xd4/\x12\x1b\n\x16SEARCH_QUERY_TOO_SHORT\x10\xd5/\x12*\n%MANAGED_RESOURCE_GROUP_DOES_NOT_EXIST\x10\xd9\x36\x12\x1e\n\x19PERMISSION_NOT_PROPAGATED\x10\xda\x36\x12\x17\n\x12\x44\x45PLOYMENT_TIMEOUT\x10\xdb\x36\x12\x11\n\x0cGIT_CONFLICT\x10\xc1>\x12\x14\n\x0fGIT_UNKNOWN_REF\x10\xc2>\x12!\n\x1cGIT_SENSITIVE_TOKEN_DETECTED\x10\xc3>\x12\x1e\n\x19GIT_URL_NOT_ON_ALLOW_LIST\x10\xc4>\x12\x15\n\x10GIT_REMOTE_ERROR\x10\xc5>\x12\x1f\n\x1aPROJECTS_OPERATION_TIMEOUT\x10\xc6>\x12\x17\n\x12IPYNB_FILE_IN_REPO\x10\xc7>\x12\x1e\n\x19INSECURE_PARTNER_RESPONSE\x10\xa4?\x12\x1f\n\x1aMALFORMED_PARTNER_RESPONSE\x10\xa5?\x12\x1d\n\x18METASTORE_DOES_NOT_EXIST\x10\xa8\x46\x12\x17\n\x12\x44\x41\x43_DOES_NOT_EXIST\x10\xa9\x46\x12\x1b\n\x16\x43\x41TALOG_DOES_NOT_EXIST\x10\xaa\x46\x12\x1a\n\x15SCHEMA_DOES_NOT_EXIST\x10\xab\x46\x12\x19\n\x14TABLE_DOES_NOT_EXIST\x10\xac\x46\x12\x19\n\x14SHARE_DOES_NOT_EXIST\x10\xad\x46\x12\x1d\n\x18RECIPIENT_DOES_NOT_EXIST\x10\xae\x46\x12&\n!STORAGE_CREDENTIAL_DOES_NOT_EXIST\x10\xaf\x46\x12%\n EXTERNAL_LOCATION_DOES_NOT_EXIST\x10\xb0\x46\x12\x1d\n\x18PRINCIPAL_DOES_NOT_EXIST\x10\xb1\x46\x12\x1c\n\x17PROVIDER_DOES_NOT_EXIST\x10\xb2\x46\x12\x1d\n\x18METASTORE_ALREADY_EXISTS\x10\xbc\x46\x12\x17\n\x12\x44\x41\x43_ALREADY_EXISTS\x10\xbd\x46\x12\x1b\n\x16\x43\x41TALOG_ALREADY_EXISTS\x10\xbe\x46\x12\x1a\n\x15SCHEMA_ALREADY_EXISTS\x10\xbf\x46\x12\x19\n\x14TABLE_ALREADY_EXISTS\x10\xc0\x46\x12\x19\n\x14SHARE_ALREADY_EXISTS\x10\xc1\x46\x12\x1d\n\x18RECIPIENT_ALREADY_EXISTS\x10\xc2\x46\x12&\n!STORAGE_CREDENTIAL_ALREADY_EXISTS\x10\xc3\x46\x12%\n EXTERNAL_LOCATION_ALREADY_EXISTS\x10\xc4\x46\x12\x1c\n\x17PROVIDER_ALREADY_EXISTS\x10\xc5\x46\x12\x16\n\x11\x43\x41TALOG_NOT_EMPTY\x10\xd0\x46\x12\x15\n\x10SCHEMA_NOT_EMPTY\x10\xd1\x46\x12\x18\n\x13METASTORE_NOT_EMPTY\x10\xd2\x46\x12\"\n\x1dPROVIDER_SHARE_NOT_ACCESSIBLE\x10\xe4\x46:G\n\nvisibility\x12\x1d.google.protobuf.FieldOptions\x18\xee\x90\x03 \x01(\x0e\x32\x12.mlflow.Visibility::\n\x11validate_required\x12\x1d.google.protobuf.FieldOptions\x18\xef\x90\x03 \x01(\x08:4\n\x0bjson_inline\x12\x1d.google.protobuf.FieldOptions\x18\xf0\x90\x03 \x01(\x08:1\n\x08json_map\x12\x1d.google.protobuf.FieldOptions\x18\xf1\x90\x03 \x01(\x08:Q\n\tfield_doc\x12\x1d.google.protobuf.FieldOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadata:K\n\x03rpc\x12\x1e.google.protobuf.MethodOptions\x18\xee\x90\x03 \x01(\x0b\x32\x1c.mlflow.DatabricksRpcOptions:S\n\nmethod_doc\x12\x1e.google.protobuf.MethodOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadata:S\n\x07graphql\x12\x1e.google.protobuf.MethodOptions\x18\xc7\x91\x03 \x01(\x0b\x32 .mlflow.DatabricksGraphqlOptions:U\n\x0bmessage_doc\x12\x1f.google.protobuf.MessageOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadata:U\n\x0bservice_doc\x12\x1f.google.protobuf.ServiceOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadata:O\n\x08\x65num_doc\x12\x1c.google.protobuf.EnumOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadata:V\n\x15\x65num_value_visibility\x12!.google.protobuf.EnumValueOptions\x18\xee\x90\x03 \x01(\x0e\x32\x12.mlflow.Visibility:Z\n\x0e\x65num_value_doc\x12!.google.protobuf.EnumValueOptions\x18\xf2\x90\x03 \x03(\x0b\x32\x1d.mlflow.DocumentationMetadataB*\n#com.databricks.api.proto.databricks\xe2?\x02\x10\x01')

  _VISIBILITY = DESCRIPTOR.enum_types_by_name['Visibility']
  Visibility = enum_type_wrapper.EnumTypeWrapper(_VISIBILITY)
  _ERRORCODE = DESCRIPTOR.enum_types_by_name['ErrorCode']
  ErrorCode = enum_type_wrapper.EnumTypeWrapper(_ERRORCODE)
  PUBLIC = 1
  INTERNAL = 2
  PUBLIC_UNDOCUMENTED = 3
  INTERNAL_ERROR = 1
  TEMPORARILY_UNAVAILABLE = 2
  IO_ERROR = 3
  BAD_REQUEST = 4
  SERVICE_UNDER_MAINTENANCE = 5
  WORKSPACE_TEMPORARILY_UNAVAILABLE = 6
  DEADLINE_EXCEEDED = 7
  CANCELLED = 8
  RESOURCE_EXHAUSTED = 9
  ABORTED = 10
  NOT_FOUND = 11
  ALREADY_EXISTS = 12
  UNAUTHENTICATED = 13
  INVALID_PARAMETER_VALUE = 1000
  ENDPOINT_NOT_FOUND = 1001
  MALFORMED_REQUEST = 1002
  INVALID_STATE = 1003
  PERMISSION_DENIED = 1004
  FEATURE_DISABLED = 1005
  CUSTOMER_UNAUTHORIZED = 1006
  REQUEST_LIMIT_EXCEEDED = 1007
  RESOURCE_CONFLICT = 1008
  UNPARSEABLE_HTTP_ERROR = 1009
  NOT_IMPLEMENTED = 1010
  DATA_LOSS = 1011
  INVALID_STATE_TRANSITION = 2001
  COULD_NOT_ACQUIRE_LOCK = 2002
  RESOURCE_ALREADY_EXISTS = 3001
  RESOURCE_DOES_NOT_EXIST = 3002
  QUOTA_EXCEEDED = 4001
  MAX_BLOCK_SIZE_EXCEEDED = 4002
  MAX_READ_SIZE_EXCEEDED = 4003
  PARTIAL_DELETE = 4004
  MAX_LIST_SIZE_EXCEEDED = 4005
  DRY_RUN_FAILED = 5001
  RESOURCE_LIMIT_EXCEEDED = 5002
  DIRECTORY_NOT_EMPTY = 6001
  DIRECTORY_PROTECTED = 6002
  MAX_NOTEBOOK_SIZE_EXCEEDED = 6003
  MAX_CHILD_NODE_SIZE_EXCEEDED = 6004
  SEARCH_QUERY_TOO_LONG = 6100
  SEARCH_QUERY_TOO_SHORT = 6101
  MANAGED_RESOURCE_GROUP_DOES_NOT_EXIST = 7001
  PERMISSION_NOT_PROPAGATED = 7002
  DEPLOYMENT_TIMEOUT = 7003
  GIT_CONFLICT = 8001
  GIT_UNKNOWN_REF = 8002
  GIT_SENSITIVE_TOKEN_DETECTED = 8003
  GIT_URL_NOT_ON_ALLOW_LIST = 8004
  GIT_REMOTE_ERROR = 8005
  PROJECTS_OPERATION_TIMEOUT = 8006
  IPYNB_FILE_IN_REPO = 8007
  INSECURE_PARTNER_RESPONSE = 8100
  MALFORMED_PARTNER_RESPONSE = 8101
  METASTORE_DOES_NOT_EXIST = 9000
  DAC_DOES_NOT_EXIST = 9001
  CATALOG_DOES_NOT_EXIST = 9002
  SCHEMA_DOES_NOT_EXIST = 9003
  TABLE_DOES_NOT_EXIST = 9004
  SHARE_DOES_NOT_EXIST = 9005
  RECIPIENT_DOES_NOT_EXIST = 9006
  STORAGE_CREDENTIAL_DOES_NOT_EXIST = 9007
  EXTERNAL_LOCATION_DOES_NOT_EXIST = 9008
  PRINCIPAL_DOES_NOT_EXIST = 9009
  PROVIDER_DOES_NOT_EXIST = 9010
  METASTORE_ALREADY_EXISTS = 9020
  DAC_ALREADY_EXISTS = 9021
  CATALOG_ALREADY_EXISTS = 9022
  SCHEMA_ALREADY_EXISTS = 9023
  TABLE_ALREADY_EXISTS = 9024
  SHARE_ALREADY_EXISTS = 9025
  RECIPIENT_ALREADY_EXISTS = 9026
  STORAGE_CREDENTIAL_ALREADY_EXISTS = 9027
  EXTERNAL_LOCATION_ALREADY_EXISTS = 9028
  PROVIDER_ALREADY_EXISTS = 9029
  CATALOG_NOT_EMPTY = 9040
  SCHEMA_NOT_EMPTY = 9041
  METASTORE_NOT_EMPTY = 9042
  PROVIDER_SHARE_NOT_ACCESSIBLE = 9060

  VISIBILITY_FIELD_NUMBER = 51310
  visibility = DESCRIPTOR.extensions_by_name['visibility']
  VALIDATE_REQUIRED_FIELD_NUMBER = 51311
  validate_required = DESCRIPTOR.extensions_by_name['validate_required']
  JSON_INLINE_FIELD_NUMBER = 51312
  json_inline = DESCRIPTOR.extensions_by_name['json_inline']
  JSON_MAP_FIELD_NUMBER = 51313
  json_map = DESCRIPTOR.extensions_by_name['json_map']
  FIELD_DOC_FIELD_NUMBER = 51314
  field_doc = DESCRIPTOR.extensions_by_name['field_doc']
  RPC_FIELD_NUMBER = 51310
  rpc = DESCRIPTOR.extensions_by_name['rpc']
  METHOD_DOC_FIELD_NUMBER = 51314
  method_doc = DESCRIPTOR.extensions_by_name['method_doc']
  GRAPHQL_FIELD_NUMBER = 51399
  graphql = DESCRIPTOR.extensions_by_name['graphql']
  MESSAGE_DOC_FIELD_NUMBER = 51314
  message_doc = DESCRIPTOR.extensions_by_name['message_doc']
  SERVICE_DOC_FIELD_NUMBER = 51314
  service_doc = DESCRIPTOR.extensions_by_name['service_doc']
  ENUM_DOC_FIELD_NUMBER = 51314
  enum_doc = DESCRIPTOR.extensions_by_name['enum_doc']
  ENUM_VALUE_VISIBILITY_FIELD_NUMBER = 51310
  enum_value_visibility = DESCRIPTOR.extensions_by_name['enum_value_visibility']
  ENUM_VALUE_DOC_FIELD_NUMBER = 51314
  enum_value_doc = DESCRIPTOR.extensions_by_name['enum_value_doc']

  _DATABRICKSRPCOPTIONS = DESCRIPTOR.message_types_by_name['DatabricksRpcOptions']
  _DATABRICKSGRAPHQLOPTIONS = DESCRIPTOR.message_types_by_name['DatabricksGraphqlOptions']
  _HTTPENDPOINT = DESCRIPTOR.message_types_by_name['HttpEndpoint']
  _APIVERSION = DESCRIPTOR.message_types_by_name['ApiVersion']
  _RATELIMIT = DESCRIPTOR.message_types_by_name['RateLimit']
  _DOCUMENTATIONMETADATA = DESCRIPTOR.message_types_by_name['DocumentationMetadata']
  DatabricksRpcOptions = _reflection.GeneratedProtocolMessageType('DatabricksRpcOptions', (_message.Message,), {
    'DESCRIPTOR' : _DATABRICKSRPCOPTIONS,
    '__module__' : 'databricks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DatabricksRpcOptions)
    })
  _sym_db.RegisterMessage(DatabricksRpcOptions)

  DatabricksGraphqlOptions = _reflection.GeneratedProtocolMessageType('DatabricksGraphqlOptions', (_message.Message,), {
    'DESCRIPTOR' : _DATABRICKSGRAPHQLOPTIONS,
    '__module__' : 'databricks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DatabricksGraphqlOptions)
    })
  _sym_db.RegisterMessage(DatabricksGraphqlOptions)

  HttpEndpoint = _reflection.GeneratedProtocolMessageType('HttpEndpoint', (_message.Message,), {
    'DESCRIPTOR' : _HTTPENDPOINT,
    '__module__' : 'databricks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.HttpEndpoint)
    })
  _sym_db.RegisterMessage(HttpEndpoint)

  ApiVersion = _reflection.GeneratedProtocolMessageType('ApiVersion', (_message.Message,), {
    'DESCRIPTOR' : _APIVERSION,
    '__module__' : 'databricks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.ApiVersion)
    })
  _sym_db.RegisterMessage(ApiVersion)

  RateLimit = _reflection.GeneratedProtocolMessageType('RateLimit', (_message.Message,), {
    'DESCRIPTOR' : _RATELIMIT,
    '__module__' : 'databricks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.RateLimit)
    })
  _sym_db.RegisterMessage(RateLimit)

  DocumentationMetadata = _reflection.GeneratedProtocolMessageType('DocumentationMetadata', (_message.Message,), {
    'DESCRIPTOR' : _DOCUMENTATIONMETADATA,
    '__module__' : 'databricks_pb2'
    # @@protoc_insertion_point(class_scope:mlflow.DocumentationMetadata)
    })
  _sym_db.RegisterMessage(DocumentationMetadata)

  if _descriptor._USE_C_DESCRIPTORS == False:
    google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(visibility)
    google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(validate_required)
    google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(json_inline)
    google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(json_map)
    google_dot_protobuf_dot_descriptor__pb2.FieldOptions.RegisterExtension(field_doc)
    google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(rpc)
    google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(method_doc)
    google_dot_protobuf_dot_descriptor__pb2.MethodOptions.RegisterExtension(graphql)
    google_dot_protobuf_dot_descriptor__pb2.MessageOptions.RegisterExtension(message_doc)
    google_dot_protobuf_dot_descriptor__pb2.ServiceOptions.RegisterExtension(service_doc)
    google_dot_protobuf_dot_descriptor__pb2.EnumOptions.RegisterExtension(enum_doc)
    google_dot_protobuf_dot_descriptor__pb2.EnumValueOptions.RegisterExtension(enum_value_visibility)
    google_dot_protobuf_dot_descriptor__pb2.EnumValueOptions.RegisterExtension(enum_value_doc)

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b'\n#com.databricks.api.proto.databricks\342?\002\020\001'
    _VISIBILITY._serialized_start=668
    _VISIBILITY._serialized_end=731
    _ERRORCODE._serialized_start=734
    _ERRORCODE._serialized_end=2907
    _DATABRICKSRPCOPTIONS._serialized_start=86
    _DATABRICKSRPCOPTIONS._serialized_end=291
    _DATABRICKSGRAPHQLOPTIONS._serialized_start=293
    _DATABRICKSGRAPHQLOPTIONS._serialized_end=319
    _HTTPENDPOINT._serialized_start=321
    _HTTPENDPOINT._serialized_end=406
    _APIVERSION._serialized_start=408
    _APIVERSION._serialized_end=450
    _RATELIMIT._serialized_start=452
    _RATELIMIT._serialized_end=516
    _DOCUMENTATIONMETADATA._serialized_start=519
    _DOCUMENTATIONMETADATA._serialized_end=666
  # @@protoc_insertion_point(module_scope)




from __future__ import annotations

import json

from deepeval.models import LiteLLMModel
from deepeval.models.base_model import DeepEvalBaseLLM
from pydantic import ValidationError

from mlflow.exceptions import MlflowException
from mlflow.genai.judges.adapters.databricks_managed_judge_adapter import (
    call_chat_completions,
)
from mlflow.genai.judges.adapters.databricks_serving_endpoint_adapter import (
    _invoke_databricks_serving_endpoint,
)
from mlflow.genai.judges.constants import _DATABRICKS_DEFAULT_JUDGE_MODEL


def _build_json_prompt_with_schema(prompt: str, schema) -> str:
    return (
        f"{prompt}\n\n"
        f"IMPORTANT: Return your response as valid JSON matching this schema: "
        f"{schema.model_json_schema()}\n"
        f"Return ONLY the JSON object, no additional text or markdown formatting."
    )


def _parse_json_output_with_schema(output: str, schema):
    try:
        json_data = json.loads(output)
    except json.JSONDecodeError as e:
        raise ValueError(f"Failed to parse JSON output: {e}\nOutput: {output}")

    try:
        return schema(**json_data)
    except ValidationError as e:
        raise ValueError(f"Failed to validate output against schema: {e}\nOutput: {output}")
    except TypeError as e:
        raise ValueError(f"Failed to instantiate schema with data: {e}\nOutput: {output}")


class DatabricksDeepEvalLLM(DeepEvalBaseLLM):
    """
    DeepEval model adapter for Databricks managed judge.

    Uses the default Databricks endpoint via call_chat_completions.
    """

    def __init__(self):
        super().__init__(model_name=_DATABRICKS_DEFAULT_JUDGE_MODEL)

    def load_model(self, **kwargs):
        return self

    def generate(self, prompt: str, schema=None) -> str:
        if schema is not None:
            # TODO: Add support for structured outputs once the Databricks endpoint supports it
            json_prompt = _build_json_prompt_with_schema(prompt, schema)
            result = call_chat_completions(user_prompt=json_prompt, system_prompt="")
            return _parse_json_output_with_schema(result.output.strip(), schema)
        else:
            result = call_chat_completions(user_prompt=prompt, system_prompt="")
            return result.output

    async def a_generate(self, prompt: str, schema=None) -> str:
        return self.generate(prompt, schema=schema)

    def get_model_name(self) -> str:
        return _DATABRICKS_DEFAULT_JUDGE_MODEL


class DatabricksServingEndpointDeepEvalLLM(DeepEvalBaseLLM):
    """
    DeepEval model adapter for Databricks serving endpoints.

    Uses the model serving API via _invoke_databricks_serving_endpoint.
    """

    def __init__(self, endpoint_name: str):
        self._endpoint_name = endpoint_name
        super().__init__(model_name=f"databricks:/{endpoint_name}")

    def load_model(self, **kwargs):
        return self

    def generate(self, prompt: str, schema=None) -> str:
        if schema is not None:
            # TODO: Use response_format parameter once Databricks serving endpoints support it
            json_prompt = _build_json_prompt_with_schema(prompt, schema)
            output = _invoke_databricks_serving_endpoint(
                model_name=self._endpoint_name,
                prompt=json_prompt,
                num_retries=3,
                response_format=None,
            )
            return _parse_json_output_with_schema(output.response, schema)
        else:
            output = _invoke_databricks_serving_endpoint(
                model_name=self._endpoint_name,
                prompt=prompt,
                num_retries=3,
                response_format=None,
            )
            return output.response

    async def a_generate(self, prompt: str, schema=None) -> str:
        return self.generate(prompt, schema=schema)

    def get_model_name(self) -> str:
        return f"databricks:/{self._endpoint_name}"


def create_deepeval_model(model_uri: str):
    if model_uri == "databricks":
        return DatabricksDeepEvalLLM()
    elif model_uri.startswith("databricks:/"):
        endpoint_name = model_uri.split(":", 1)[1].removeprefix("/")
        return DatabricksServingEndpointDeepEvalLLM(endpoint_name)
    elif ":" in model_uri:
        # LiteLLM model format with provider: provider:/model_name (e.g., openai:/gpt-4)
        provider, model_name = model_uri.split(":", 1)
        model_name = model_name.removeprefix("/")
        return LiteLLMModel(model=f"{provider}/{model_name}")
    else:
        raise MlflowException.invalid_parameter_value(
            f"Invalid model_uri format: '{model_uri}'. "
            f"Must be 'databricks' or include a provider prefix (e.g., 'openai:/gpt-4')."
        )



from __future__ import annotations

from typing import ClassVar

from mlflow.genai.judges.builtin import _MODEL_API_DOC
from mlflow.genai.scorers.ragas import RagasScorer
from mlflow.utils.annotations import experimental
from mlflow.utils.docstring_utils import format_docstring


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class ContextPrecision(RagasScorer):
    """
    Evaluates whether relevant nodes in the retrieval context are ranked higher than
    irrelevant ones.

    Args:
        model: {{ model }}
        **metric_kwargs: Additional metric-specific parameters

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.ragas import ContextPrecision

            scorer = ContextPrecision(model="openai:/gpt-4")
            feedback = scorer(trace=trace)
    """

    metric_name: ClassVar[str] = "ContextPrecision"


@experimental(version="3.8.0")
class NonLLMContextPrecisionWithReference(RagasScorer):
    """
    Deterministic metric that evaluates context precision using non-LLM methods using expectations.

    Args:
        **metric_kwargs: Additional metric-specific parameters

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.ragas import NonLLMContextPrecisionWithReference

            scorer = NonLLMContextPrecisionWithReference()
            feedback = scorer(trace=trace)
    """

    metric_name: ClassVar[str] = "NonLLMContextPrecisionWithReference"

    def __init__(self, **metric_kwargs):
        self._validate_kwargs(**metric_kwargs)
        super().__init__(metric_name=self.metric_name, model=None, **metric_kwargs)


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class ContextRecall(RagasScorer):
    """
    Evaluates whether the retrieval context contains all necessary information.

    Args:
        model: {{ model }}
        **metric_kwargs: Additional metric-specific parameters

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.ragas import ContextRecall

            scorer = ContextRecall(model="openai:/gpt-4")
            feedback = scorer(trace=trace)
    """

    metric_name: ClassVar[str] = "ContextRecall"


@experimental(version="3.8.0")
class NonLLMContextRecall(RagasScorer):
    """
    Deterministic metric that evaluates context recall without using an LLM.

    Args:
        **metric_kwargs: Additional metric-specific parameters

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.ragas import NonLLMContextRecall

            scorer = NonLLMContextRecall()
            feedback = scorer(trace=trace)
    """

    metric_name: ClassVar[str] = "NonLLMContextRecall"

    def __init__(self, **metric_kwargs):
        self._validate_kwargs(**metric_kwargs)
        super().__init__(metric_name=self.metric_name, model=None, **metric_kwargs)


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class ContextEntityRecall(RagasScorer):
    """
    Evaluates entity recall in the retrieval context.

    Args:
        model: {{ model }}
        **metric_kwargs: Additional metric-specific parameters

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.ragas import ContextEntityRecall

            scorer = ContextEntityRecall(model="openai:/gpt-4")
            feedback = scorer(trace=trace)
    """

    metric_name: ClassVar[str] = "ContextEntityRecall"


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class NoiseSensitivity(RagasScorer):
    """
    Evaluates how sensitive the model is to noise in the retrieval context.

    Args:
        model: {{ model }}
        **metric_kwargs: Additional metric-specific parameters

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.ragas import NoiseSensitivity

            scorer = NoiseSensitivity(model="openai:/gpt-4")
            feedback = scorer(trace=trace)
    """

    metric_name: ClassVar[str] = "NoiseSensitivity"


@experimental(version="3.8.0")
@format_docstring(_MODEL_API_DOC)
class Faithfulness(RagasScorer):
    """
    Evaluates whether the output is factually consistent with the retrieval context.

    Args:
        model: {{ model }}
        **metric_kwargs: Additional metric-specific parameters

    Examples:
        .. code-block:: python

            from mlflow.genai.scorers.ragas import Faithfulness

            scorer = Faithfulness(model="openai:/gpt-4")
            feedback = scorer(trace=trace)
    """

    metric_name: ClassVar[str] = "Faithfulness"



import importlib
import inspect
import json
import logging
from typing import Any, Callable

from llama_index.core import PromptTemplate
from llama_index.core.base.embeddings.base import BaseEmbedding
from llama_index.core.callbacks.base import CallbackManager
from llama_index.core.schema import BaseComponent

_logger = logging.getLogger(__name__)


def _get_object_import_path(o: object) -> str:
    if not inspect.isclass(o):
        o = o.__class__

    module_name = inspect.getmodule(o).__name__
    class_name = o.__qualname__

    # Validate the import
    module = importlib.import_module(module_name)
    if not hasattr(module, class_name):
        raise ValueError(f"Module {module} does not have {class_name}")

    return f"{module_name}.{class_name}"


def _sanitize_api_key(object_as_dict: dict[str, str]) -> dict[str, str]:
    return {k: v for k, v in object_as_dict.items() if "api_key" not in k.lower()}


def object_to_dict(o: object):
    if isinstance(o, (list, tuple)):
        return [object_to_dict(v) for v in o]

    if isinstance(o, BaseComponent):
        # we can't serialize callables in the model fields
        callable_fields = set()
        # Access model_fields from the class to avoid pydantic deprecation warning
        fields = (
            o.__class__.model_fields if hasattr(o.__class__, "model_fields") else o.model_fields
        )
        for k, v in fields.items():
            field_val = getattr(o, k, None)
            # Exclude all callable fields, including those with default values
            # to prevent serialization issues in llama_index
            if callable(field_val):
                callable_fields.add(k)
        # exclude default values from serialization to avoid
        # unnecessary clutter in the serialized object
        o_state_as_dict = o.to_dict(exclude=callable_fields)

        if o_state_as_dict != {}:
            o_state_as_dict = _sanitize_api_key(o_state_as_dict)
            o_state_as_dict.pop("class_name")
        else:
            return o_state_as_dict

        return {
            "object_constructor": _get_object_import_path(o),
            "object_kwargs": o_state_as_dict,
        }
    else:
        return None


def _construct_prompt_template_object(
    constructor: Callable[..., PromptTemplate], kwargs: dict[str, Any]
) -> PromptTemplate:
    """Construct a PromptTemplate object based on the constructor and kwargs.

    This method is necessary because the `template_vars` cannot be passed directly to the
    constructor and needs to be set on an instantiated object.
    """
    if template := kwargs.pop("template", None):
        prompt_template = constructor(template)
        for k, v in kwargs.items():
            setattr(prompt_template, k, v)

        return prompt_template
    else:
        raise ValueError(
            "'template' is a required kwargs and is not present in the prompt template kwargs."
        )


def dict_to_object(object_representation: dict[str, Any]) -> object:
    if "object_constructor" not in object_representation:
        raise ValueError("'object_constructor' key not found in dict.")
    if "object_kwargs" not in object_representation:
        raise ValueError("'object_kwargs' key not found in dict.")

    constructor_str = object_representation["object_constructor"]
    kwargs = object_representation["object_kwargs"]

    import_path, class_name = constructor_str.rsplit(".", 1)
    module = importlib.import_module(import_path)

    if isinstance(module, PromptTemplate):
        return _construct_prompt_template_object(module, kwargs)
    else:
        object_class = getattr(module, class_name)

        # Many embeddings model accepts parameter `model`, while BaseEmbedding accepts `model_name`.
        # Both parameters will be serialized as kwargs, but passing both to the constructor will
        # raise duplicate argument error. Some class like OpenAIEmbedding handles this in its
        # constructor, but not all integrations do. Therefore, we have to handle it here.
        # E.g. https://github.com/run-llama/llama_index/blob/2b18eb4654b14c68d63f6239cddb10740668fbc8/llama-index-integrations/embeddings/llama-index-embeddings-openai/llama_index/embeddings/openai/base.py#L316-L320
        if (
            issubclass(object_class, BaseEmbedding)
            and (model := kwargs.get("model"))
            and (model_name := kwargs.get("model_name"))
            and model == model_name
        ):
            kwargs.pop("model_name")

        return object_class.from_dict(kwargs)


def _deserialize_dict_of_objects(path: str) -> dict[str, Any]:
    with open(path) as f:
        to_deserialize = json.load(f)

        output = {}
        for k, v in to_deserialize.items():
            if isinstance(v, list):
                output.update({k: [dict_to_object(vv) for vv in v]})
            else:
                output.update({k: dict_to_object(v)})

        return output


def serialize_settings(path: str) -> None:
    """Serialize the global LlamaIndex Settings object to a JSON file at the given path."""
    from llama_index.core import Settings

    _logger.info(
        "API key(s) will be removed from the global Settings object during serialization "
        "to protect against key leakage. At inference time, the key(s) must be passed as "
        "environment variables."
    )

    to_serialize = {}
    unsupported_objects = []

    for k, v in Settings.__dict__.items():
        if v is None:
            continue

        # Setting.callback_manager is default to an empty CallbackManager instance.
        if (k == "_callback_manager") and isinstance(v, CallbackManager) and v.handlers == []:
            continue

        def _convert(obj):
            object_json = object_to_dict(obj)
            if object_json is None:
                prop_name = k.removeprefix("_")
                unsupported_objects.append((prop_name, v))
            return object_json

        if isinstance(v, list):
            to_serialize[k] = [_convert(obj) for obj in v if v is not None]
        else:
            if (object_json := _convert(v)) and (object_json is not None):
                to_serialize[k] = object_json

    if unsupported_objects:
        msg = (
            "The following objects in Settings are not supported for serialization and will not "
            "be logged with your model. MLflow only supports serialization of objects that inherit "
            "from llama_index.core.schema.BaseComponent.\n"
        )
        msg += "\n".join(f" - {type(v).__name__} for Settings.{k}" for k, v in unsupported_objects)
        _logger.info(msg)

    with open(path, "w") as f:
        json.dump(to_serialize, f, indent=2)


def deserialize_settings(path: str):
    """Deserialize the global LlamaIndex Settings object from a JSON file at the given path."""
    settings_dict = _deserialize_dict_of_objects(path)

    from llama_index.core import Settings

    for k, v in settings_dict.items():
        # To use the property setter rather than directly setting the private attribute e.g. _llm
        k = k.removeprefix("_")
        setattr(Settings, k, v)



import assessments_pb2 as _assessments_pb2
import databricks_pb2 as _databricks_pb2
from google.protobuf import duration_pb2 as _duration_pb2
from google.protobuf import field_mask_pb2 as _field_mask_pb2
from google.protobuf import timestamp_pb2 as _timestamp_pb2
from opentelemetry.proto.trace.v1 import trace_pb2 as _trace_pb2
from scalapb import scalapb_pb2 as _scalapb_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import service as _service
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class UCSchemaLocation(_message.Message):
    __slots__ = ("catalog_name", "schema_name", "otel_spans_table_name", "otel_logs_table_name")
    CATALOG_NAME_FIELD_NUMBER: _ClassVar[int]
    SCHEMA_NAME_FIELD_NUMBER: _ClassVar[int]
    OTEL_SPANS_TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    OTEL_LOGS_TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    catalog_name: str
    schema_name: str
    otel_spans_table_name: str
    otel_logs_table_name: str
    def __init__(self, catalog_name: _Optional[str] = ..., schema_name: _Optional[str] = ..., otel_spans_table_name: _Optional[str] = ..., otel_logs_table_name: _Optional[str] = ...) -> None: ...

class MlflowExperimentLocation(_message.Message):
    __slots__ = ("experiment_id",)
    EXPERIMENT_ID_FIELD_NUMBER: _ClassVar[int]
    experiment_id: str
    def __init__(self, experiment_id: _Optional[str] = ...) -> None: ...

class InferenceTableLocation(_message.Message):
    __slots__ = ("full_table_name",)
    FULL_TABLE_NAME_FIELD_NUMBER: _ClassVar[int]
    full_table_name: str
    def __init__(self, full_table_name: _Optional[str] = ...) -> None: ...

class TraceLocation(_message.Message):
    __slots__ = ("type", "mlflow_experiment", "inference_table", "uc_schema")
    class TraceLocationType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        TRACE_LOCATION_TYPE_UNSPECIFIED: _ClassVar[TraceLocation.TraceLocationType]
        MLFLOW_EXPERIMENT: _ClassVar[TraceLocation.TraceLocationType]
        INFERENCE_TABLE: _ClassVar[TraceLocation.TraceLocationType]
        UC_SCHEMA: _ClassVar[TraceLocation.TraceLocationType]
    TRACE_LOCATION_TYPE_UNSPECIFIED: TraceLocation.TraceLocationType
    MLFLOW_EXPERIMENT: TraceLocation.TraceLocationType
    INFERENCE_TABLE: TraceLocation.TraceLocationType
    UC_SCHEMA: TraceLocation.TraceLocationType
    TYPE_FIELD_NUMBER: _ClassVar[int]
    MLFLOW_EXPERIMENT_FIELD_NUMBER: _ClassVar[int]
    INFERENCE_TABLE_FIELD_NUMBER: _ClassVar[int]
    UC_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    type: TraceLocation.TraceLocationType
    mlflow_experiment: MlflowExperimentLocation
    inference_table: InferenceTableLocation
    uc_schema: UCSchemaLocation
    def __init__(self, type: _Optional[_Union[TraceLocation.TraceLocationType, str]] = ..., mlflow_experiment: _Optional[_Union[MlflowExperimentLocation, _Mapping]] = ..., inference_table: _Optional[_Union[InferenceTableLocation, _Mapping]] = ..., uc_schema: _Optional[_Union[UCSchemaLocation, _Mapping]] = ...) -> None: ...

class TraceInfo(_message.Message):
    __slots__ = ("trace_id", "client_request_id", "trace_location", "request_preview", "response_preview", "request_time", "execution_duration", "state", "trace_metadata", "assessments", "tags")
    class State(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        STATE_UNSPECIFIED: _ClassVar[TraceInfo.State]
        OK: _ClassVar[TraceInfo.State]
        ERROR: _ClassVar[TraceInfo.State]
        IN_PROGRESS: _ClassVar[TraceInfo.State]
    STATE_UNSPECIFIED: TraceInfo.State
    OK: TraceInfo.State
    ERROR: TraceInfo.State
    IN_PROGRESS: TraceInfo.State
    class TraceMetadataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    class TagsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    CLIENT_REQUEST_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_LOCATION_FIELD_NUMBER: _ClassVar[int]
    REQUEST_PREVIEW_FIELD_NUMBER: _ClassVar[int]
    RESPONSE_PREVIEW_FIELD_NUMBER: _ClassVar[int]
    REQUEST_TIME_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_DURATION_FIELD_NUMBER: _ClassVar[int]
    STATE_FIELD_NUMBER: _ClassVar[int]
    TRACE_METADATA_FIELD_NUMBER: _ClassVar[int]
    ASSESSMENTS_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    trace_id: str
    client_request_id: str
    trace_location: TraceLocation
    request_preview: str
    response_preview: str
    request_time: _timestamp_pb2.Timestamp
    execution_duration: _duration_pb2.Duration
    state: TraceInfo.State
    trace_metadata: _containers.ScalarMap[str, str]
    assessments: _containers.RepeatedCompositeFieldContainer[Assessment]
    tags: _containers.ScalarMap[str, str]
    def __init__(self, trace_id: _Optional[str] = ..., client_request_id: _Optional[str] = ..., trace_location: _Optional[_Union[TraceLocation, _Mapping]] = ..., request_preview: _Optional[str] = ..., response_preview: _Optional[str] = ..., request_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., execution_duration: _Optional[_Union[_duration_pb2.Duration, _Mapping]] = ..., state: _Optional[_Union[TraceInfo.State, str]] = ..., trace_metadata: _Optional[_Mapping[str, str]] = ..., assessments: _Optional[_Iterable[_Union[Assessment, _Mapping]]] = ..., tags: _Optional[_Mapping[str, str]] = ...) -> None: ...

class CreateTraceInfo(_message.Message):
    __slots__ = ("location_id", "trace_info")
    class Response(_message.Message):
        __slots__ = ("trace_info",)
        TRACE_INFO_FIELD_NUMBER: _ClassVar[int]
        trace_info: TraceInfo
        def __init__(self, trace_info: _Optional[_Union[TraceInfo, _Mapping]] = ...) -> None: ...
    LOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_INFO_FIELD_NUMBER: _ClassVar[int]
    location_id: str
    trace_info: TraceInfo
    def __init__(self, location_id: _Optional[str] = ..., trace_info: _Optional[_Union[TraceInfo, _Mapping]] = ...) -> None: ...

class TracePath(_message.Message):
    __slots__ = ("trace_location", "trace_id")
    TRACE_LOCATION_FIELD_NUMBER: _ClassVar[int]
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    trace_location: TraceLocation
    trace_id: str
    def __init__(self, trace_location: _Optional[_Union[TraceLocation, _Mapping]] = ..., trace_id: _Optional[str] = ...) -> None: ...

class Trace(_message.Message):
    __slots__ = ("trace_info", "spans")
    TRACE_INFO_FIELD_NUMBER: _ClassVar[int]
    SPANS_FIELD_NUMBER: _ClassVar[int]
    trace_info: TraceInfo
    spans: _containers.RepeatedCompositeFieldContainer[_trace_pb2.Span]
    def __init__(self, trace_info: _Optional[_Union[TraceInfo, _Mapping]] = ..., spans: _Optional[_Iterable[_Union[_trace_pb2.Span, _Mapping]]] = ...) -> None: ...

class BatchGetTraces(_message.Message):
    __slots__ = ("location_id", "trace_ids", "sql_warehouse_id")
    class Response(_message.Message):
        __slots__ = ("traces",)
        TRACES_FIELD_NUMBER: _ClassVar[int]
        traces: _containers.RepeatedCompositeFieldContainer[Trace]
        def __init__(self, traces: _Optional[_Iterable[_Union[Trace, _Mapping]]] = ...) -> None: ...
    LOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_IDS_FIELD_NUMBER: _ClassVar[int]
    SQL_WAREHOUSE_ID_FIELD_NUMBER: _ClassVar[int]
    location_id: str
    trace_ids: _containers.RepeatedScalarFieldContainer[str]
    sql_warehouse_id: str
    def __init__(self, location_id: _Optional[str] = ..., trace_ids: _Optional[_Iterable[str]] = ..., sql_warehouse_id: _Optional[str] = ...) -> None: ...

class GetTraceInfo(_message.Message):
    __slots__ = ("trace_id", "location", "sql_warehouse_id")
    class Response(_message.Message):
        __slots__ = ("trace",)
        TRACE_FIELD_NUMBER: _ClassVar[int]
        trace: Trace
        def __init__(self, trace: _Optional[_Union[Trace, _Mapping]] = ...) -> None: ...
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    LOCATION_FIELD_NUMBER: _ClassVar[int]
    SQL_WAREHOUSE_ID_FIELD_NUMBER: _ClassVar[int]
    trace_id: str
    location: str
    sql_warehouse_id: str
    def __init__(self, trace_id: _Optional[str] = ..., location: _Optional[str] = ..., sql_warehouse_id: _Optional[str] = ...) -> None: ...

class SetTraceTag(_message.Message):
    __slots__ = ("trace_id", "location_id", "key", "value")
    class Response(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    LOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    KEY_FIELD_NUMBER: _ClassVar[int]
    VALUE_FIELD_NUMBER: _ClassVar[int]
    trace_id: str
    location_id: str
    key: str
    value: str
    def __init__(self, trace_id: _Optional[str] = ..., location_id: _Optional[str] = ..., key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...

class DeleteTraceTag(_message.Message):
    __slots__ = ("trace_id", "location_id", "key", "sql_warehouse_id")
    class Response(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    LOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    KEY_FIELD_NUMBER: _ClassVar[int]
    SQL_WAREHOUSE_ID_FIELD_NUMBER: _ClassVar[int]
    trace_id: str
    location_id: str
    key: str
    sql_warehouse_id: str
    def __init__(self, trace_id: _Optional[str] = ..., location_id: _Optional[str] = ..., key: _Optional[str] = ..., sql_warehouse_id: _Optional[str] = ...) -> None: ...

class SearchTraces(_message.Message):
    __slots__ = ("locations", "filter", "max_results", "order_by", "sql_warehouse_id", "page_token")
    class Response(_message.Message):
        __slots__ = ("trace_infos", "next_page_token")
        TRACE_INFOS_FIELD_NUMBER: _ClassVar[int]
        NEXT_PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
        trace_infos: _containers.RepeatedCompositeFieldContainer[TraceInfo]
        next_page_token: str
        def __init__(self, trace_infos: _Optional[_Iterable[_Union[TraceInfo, _Mapping]]] = ..., next_page_token: _Optional[str] = ...) -> None: ...
    LOCATIONS_FIELD_NUMBER: _ClassVar[int]
    FILTER_FIELD_NUMBER: _ClassVar[int]
    MAX_RESULTS_FIELD_NUMBER: _ClassVar[int]
    ORDER_BY_FIELD_NUMBER: _ClassVar[int]
    SQL_WAREHOUSE_ID_FIELD_NUMBER: _ClassVar[int]
    PAGE_TOKEN_FIELD_NUMBER: _ClassVar[int]
    locations: _containers.RepeatedCompositeFieldContainer[TraceLocation]
    filter: str
    max_results: int
    order_by: _containers.RepeatedScalarFieldContainer[str]
    sql_warehouse_id: str
    page_token: str
    def __init__(self, locations: _Optional[_Iterable[_Union[TraceLocation, _Mapping]]] = ..., filter: _Optional[str] = ..., max_results: _Optional[int] = ..., order_by: _Optional[_Iterable[str]] = ..., sql_warehouse_id: _Optional[str] = ..., page_token: _Optional[str] = ...) -> None: ...

class CreateTraceUCStorageLocation(_message.Message):
    __slots__ = ("uc_schema", "sql_warehouse_id")
    class Response(_message.Message):
        __slots__ = ("uc_schema",)
        UC_SCHEMA_FIELD_NUMBER: _ClassVar[int]
        uc_schema: UCSchemaLocation
        def __init__(self, uc_schema: _Optional[_Union[UCSchemaLocation, _Mapping]] = ...) -> None: ...
    UC_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    SQL_WAREHOUSE_ID_FIELD_NUMBER: _ClassVar[int]
    uc_schema: UCSchemaLocation
    sql_warehouse_id: str
    def __init__(self, uc_schema: _Optional[_Union[UCSchemaLocation, _Mapping]] = ..., sql_warehouse_id: _Optional[str] = ...) -> None: ...

class LinkExperimentToUCTraceLocation(_message.Message):
    __slots__ = ("experiment_id", "uc_schema")
    class Response(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    EXPERIMENT_ID_FIELD_NUMBER: _ClassVar[int]
    UC_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    experiment_id: str
    uc_schema: UCSchemaLocation
    def __init__(self, experiment_id: _Optional[str] = ..., uc_schema: _Optional[_Union[UCSchemaLocation, _Mapping]] = ...) -> None: ...

class UnLinkExperimentToUCTraceLocation(_message.Message):
    __slots__ = ("experiment_id", "uc_schema")
    class Response(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    EXPERIMENT_ID_FIELD_NUMBER: _ClassVar[int]
    UC_SCHEMA_FIELD_NUMBER: _ClassVar[int]
    experiment_id: str
    uc_schema: UCSchemaLocation
    def __init__(self, experiment_id: _Optional[str] = ..., uc_schema: _Optional[_Union[UCSchemaLocation, _Mapping]] = ...) -> None: ...

class Assessment(_message.Message):
    __slots__ = ("assessment_id", "assessment_name", "trace_id", "trace_location", "span_id", "source", "create_time", "last_update_time", "feedback", "expectation", "rationale", "metadata", "overrides", "valid")
    class MetadataEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(self, key: _Optional[str] = ..., value: _Optional[str] = ...) -> None: ...
    ASSESSMENT_ID_FIELD_NUMBER: _ClassVar[int]
    ASSESSMENT_NAME_FIELD_NUMBER: _ClassVar[int]
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_LOCATION_FIELD_NUMBER: _ClassVar[int]
    SPAN_ID_FIELD_NUMBER: _ClassVar[int]
    SOURCE_FIELD_NUMBER: _ClassVar[int]
    CREATE_TIME_FIELD_NUMBER: _ClassVar[int]
    LAST_UPDATE_TIME_FIELD_NUMBER: _ClassVar[int]
    FEEDBACK_FIELD_NUMBER: _ClassVar[int]
    EXPECTATION_FIELD_NUMBER: _ClassVar[int]
    RATIONALE_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    OVERRIDES_FIELD_NUMBER: _ClassVar[int]
    VALID_FIELD_NUMBER: _ClassVar[int]
    assessment_id: str
    assessment_name: str
    trace_id: str
    trace_location: TraceLocation
    span_id: str
    source: _assessments_pb2.AssessmentSource
    create_time: _timestamp_pb2.Timestamp
    last_update_time: _timestamp_pb2.Timestamp
    feedback: _assessments_pb2.Feedback
    expectation: _assessments_pb2.Expectation
    rationale: str
    metadata: _containers.ScalarMap[str, str]
    overrides: str
    valid: bool
    def __init__(self, assessment_id: _Optional[str] = ..., assessment_name: _Optional[str] = ..., trace_id: _Optional[str] = ..., trace_location: _Optional[_Union[TraceLocation, _Mapping]] = ..., span_id: _Optional[str] = ..., source: _Optional[_Union[_assessments_pb2.AssessmentSource, _Mapping]] = ..., create_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., last_update_time: _Optional[_Union[_timestamp_pb2.Timestamp, _Mapping]] = ..., feedback: _Optional[_Union[_assessments_pb2.Feedback, _Mapping]] = ..., expectation: _Optional[_Union[_assessments_pb2.Expectation, _Mapping]] = ..., rationale: _Optional[str] = ..., metadata: _Optional[_Mapping[str, str]] = ..., overrides: _Optional[str] = ..., valid: bool = ...) -> None: ...

class CreateAssessment(_message.Message):
    __slots__ = ("location_id", "assessment", "sql_warehouse_id")
    class Response(_message.Message):
        __slots__ = ("assessment",)
        ASSESSMENT_FIELD_NUMBER: _ClassVar[int]
        assessment: Assessment
        def __init__(self, assessment: _Optional[_Union[Assessment, _Mapping]] = ...) -> None: ...
    LOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    ASSESSMENT_FIELD_NUMBER: _ClassVar[int]
    SQL_WAREHOUSE_ID_FIELD_NUMBER: _ClassVar[int]
    location_id: str
    assessment: Assessment
    sql_warehouse_id: str
    def __init__(self, location_id: _Optional[str] = ..., assessment: _Optional[_Union[Assessment, _Mapping]] = ..., sql_warehouse_id: _Optional[str] = ...) -> None: ...

class GetAssessment(_message.Message):
    __slots__ = ("location_id", "trace_id", "assessment_id", "sql_warehouse_id")
    class Response(_message.Message):
        __slots__ = ("assessment",)
        ASSESSMENT_FIELD_NUMBER: _ClassVar[int]
        assessment: Assessment
        def __init__(self, assessment: _Optional[_Union[Assessment, _Mapping]] = ...) -> None: ...
    LOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    ASSESSMENT_ID_FIELD_NUMBER: _ClassVar[int]
    SQL_WAREHOUSE_ID_FIELD_NUMBER: _ClassVar[int]
    location_id: str
    trace_id: str
    assessment_id: str
    sql_warehouse_id: str
    def __init__(self, location_id: _Optional[str] = ..., trace_id: _Optional[str] = ..., assessment_id: _Optional[str] = ..., sql_warehouse_id: _Optional[str] = ...) -> None: ...

class UpdateAssessment(_message.Message):
    __slots__ = ("location_id", "assessment", "update_mask", "sql_warehouse_id")
    class Response(_message.Message):
        __slots__ = ("assessment",)
        ASSESSMENT_FIELD_NUMBER: _ClassVar[int]
        assessment: Assessment
        def __init__(self, assessment: _Optional[_Union[Assessment, _Mapping]] = ...) -> None: ...
    LOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    ASSESSMENT_FIELD_NUMBER: _ClassVar[int]
    UPDATE_MASK_FIELD_NUMBER: _ClassVar[int]
    SQL_WAREHOUSE_ID_FIELD_NUMBER: _ClassVar[int]
    location_id: str
    assessment: Assessment
    update_mask: _field_mask_pb2.FieldMask
    sql_warehouse_id: str
    def __init__(self, location_id: _Optional[str] = ..., assessment: _Optional[_Union[Assessment, _Mapping]] = ..., update_mask: _Optional[_Union[_field_mask_pb2.FieldMask, _Mapping]] = ..., sql_warehouse_id: _Optional[str] = ...) -> None: ...

class DeleteAssessment(_message.Message):
    __slots__ = ("location_id", "trace_id", "assessment_id", "sql_warehouse_id")
    class Response(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    LOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_ID_FIELD_NUMBER: _ClassVar[int]
    ASSESSMENT_ID_FIELD_NUMBER: _ClassVar[int]
    SQL_WAREHOUSE_ID_FIELD_NUMBER: _ClassVar[int]
    location_id: str
    trace_id: str
    assessment_id: str
    sql_warehouse_id: str
    def __init__(self, location_id: _Optional[str] = ..., trace_id: _Optional[str] = ..., assessment_id: _Optional[str] = ..., sql_warehouse_id: _Optional[str] = ...) -> None: ...

class BatchLinkTraceToRun(_message.Message):
    __slots__ = ("location_id", "trace_ids", "run_id")
    class Response(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    LOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_IDS_FIELD_NUMBER: _ClassVar[int]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    location_id: str
    trace_ids: _containers.RepeatedScalarFieldContainer[str]
    run_id: str
    def __init__(self, location_id: _Optional[str] = ..., trace_ids: _Optional[_Iterable[str]] = ..., run_id: _Optional[str] = ...) -> None: ...

class BatchUnlinkTraceFromRun(_message.Message):
    __slots__ = ("location_id", "trace_ids", "run_id")
    class Response(_message.Message):
        __slots__ = ()
        def __init__(self) -> None: ...
    LOCATION_ID_FIELD_NUMBER: _ClassVar[int]
    TRACE_IDS_FIELD_NUMBER: _ClassVar[int]
    RUN_ID_FIELD_NUMBER: _ClassVar[int]
    location_id: str
    trace_ids: _containers.RepeatedScalarFieldContainer[str]
    run_id: str
    def __init__(self, location_id: _Optional[str] = ..., trace_ids: _Optional[_Iterable[str]] = ..., run_id: _Optional[str] = ...) -> None: ...

class DatabricksTrackingService(_service.service): ...

class DatabricksTrackingService_Stub(DatabricksTrackingService): ...



import logging
from functools import partial
from urllib.parse import parse_qs, urlparse

import mlflow
from mlflow.environment_variables import MLFLOW_SKIP_SIGNATURE_CHECK_FOR_UC_REGISTRY_MIGRATION
from mlflow.exceptions import MlflowException
from mlflow.protos.databricks_pb2 import (
    RESOURCE_ALREADY_EXISTS,
    ErrorCode,
)
from mlflow.store._unity_catalog.registry.rest_store import UcModelRegistryStore
from mlflow.store.model_registry.rest_store import RestStore
from mlflow.utils.databricks_utils import get_databricks_host_creds
from mlflow.utils.logging_utils import eprint
from mlflow.utils.uri import _DATABRICKS_UNITY_CATALOG_SCHEME

_logger = logging.getLogger(__name__)


def _extract_workspace_id_from_run_link(run_link: str | None) -> str | None:
    """
    Extract workspace ID from a Databricks run link URL.

    Args:
        run_link: URL like "https://workspace.databricks.com/?o=10002#mlflow/experiments/test-exp-id/runs/runid"
            The workspace ID is the part after the ?o= in the URL, and before the #,
            e.g. 10002 in the example above.
            The run_link is only present if the run was logged in a Databricks Workspace
            different from the registry workspace.

    Returns:
        The workspace ID as a string, or None if not found or invalid
    """
    if not run_link:
        return None

    warning_msg = (
        "Unable to get model version source run's workspace ID from source run link. "
        "The source workspace ID on the destination model version will be set to the "
        "registry workspace ID."
    )

    try:
        parsed_url = urlparse(run_link)
        query_params = parse_qs(parsed_url.query)
        workspace_id_params = query_params.get("o")
        if not workspace_id_params:
            _logger.warning(warning_msg)
            return None
        workspace_id = workspace_id_params[0]
        if int(workspace_id) < 0:
            _logger.warning(warning_msg)
            return None
        return workspace_id
    except ValueError:
        _logger.warning(warning_msg)
        return None


def _raise_unsupported_method(method, message=None):
    messages = [
        f"Method '{method}' is unsupported for models in the Workspace Model Registry. "
        f"Upgrade to Models in Unity Catalog to access the latest features. You can configure "
        f"the MLflow Python client to access models in Unity Catalog by running "
        f"mlflow.set_registry_uri('databricks-uc') before accessing models.",
    ]
    if message is not None:
        messages.append(message)
    raise MlflowException(" ".join(messages))


class DatabricksWorkspaceModelRegistryRestStore(RestStore):
    def __init__(self, store_uri, tracking_uri):
        super().__init__(get_host_creds=partial(get_databricks_host_creds, store_uri))
        self.tracking_uri = tracking_uri

    def set_registered_model_alias(self, name, alias, version):
        _raise_unsupported_method(method="set_registered_model_alias")

    def delete_registered_model_alias(self, name, alias):
        _raise_unsupported_method(method="delete_registered_model_alias")

    def get_model_version_by_alias(self, name, alias):
        _raise_unsupported_method(
            method="get_model_version_by_alias",
            message="If attempting to load a model version by alias via a URI of the form "
            "'models:/model_name@alias_name', configure the MLflow client to target Unity Catalog "
            "and try again.",
        )

    def _await_model_version_creation(self, mv, await_creation_for):
        uc_hint = (
            " For faster model version creation, use Models in Unity Catalog "
            "(https://docs.databricks.com/en/machine-learning/manage-model-lifecycle/index.html)."
        )
        self._await_model_version_creation_impl(mv, await_creation_for, hint=uc_hint)

    def copy_model_version(self, src_mv, dst_name):
        """
        Copy a model version from one registered model to another as a new model version.
        This method can be used within the Databricks workspace registry to copy model versions
        between registered models, or to migrate model versions from the Databricks workspace
        registry to Unity Catalog. During the migration, signature validation can be bypassed
        by setting the `MLFLOW_SKIP_SIGNATURE_CHECK_FOR_UC_REGISTRY_MIGRATION`environment variable
        to `True`.

        Args:
            src_mv: A :py:class:`mlflow.entities.model_registry.ModelVersion` object representing
                the source model version.
            dst_name: The name of the registered model to copy the model version to. If a
                registered model with this name does not exist, it will be created.

        Returns:
            Single :py:class:`mlflow.entities.model_registry.ModelVersion` object representing
            the cloned model version.
        """
        if dst_name.count(".") == 2:
            source_uri = f"models:/{src_mv.name}/{src_mv.version}"
            try:
                local_model_dir = mlflow.artifacts.download_artifacts(
                    artifact_uri=source_uri,
                    tracking_uri=self.tracking_uri,
                    registry_uri="databricks",
                )
            except Exception as e:
                raise MlflowException(
                    f"Unable to download model {src_mv.name} version {src_mv.version} "
                    f"artifacts from Databricks workspace registry in order to migrate "
                    f"them to Unity Catalog. Please ensure the model version artifacts "
                    f"exist and that you can download them via "
                    f"mlflow.artifacts.download_artifacts()"
                ) from e
            uc_store = UcModelRegistryStore(
                store_uri=_DATABRICKS_UNITY_CATALOG_SCHEME,
                tracking_uri=self.tracking_uri,
            )
            try:
                create_model_response = uc_store.create_registered_model(dst_name)
                eprint(f"Successfully registered model '{create_model_response.name}'.")
            except MlflowException as e:
                if e.error_code != ErrorCode.Name(RESOURCE_ALREADY_EXISTS):
                    raise
                eprint(
                    f"Registered model '{dst_name}' already exists."
                    f" Creating a new version of this model..."
                )
            skip_signature = MLFLOW_SKIP_SIGNATURE_CHECK_FOR_UC_REGISTRY_MIGRATION.get()
            source_workspace_id = _extract_workspace_id_from_run_link(src_mv.run_link)
            return uc_store._create_model_version_with_optional_signature_validation(
                name=dst_name,
                source=source_uri,
                run_id=src_mv.run_id,
                local_model_path=local_model_dir,
                model_id=src_mv.model_id,
                bypass_signature_validation=skip_signature,
                source_workspace_id=source_workspace_id,
            )
        else:
            return super().copy_model_version(src_mv, dst_name)
